! Main program to add border to the sides of the model for the absorbing boundary condition
program add_border

  use sep

  implicit none

  logical                               :: top
  integer                               :: nzin,nxin,nzout,nxout,iz,ix,abc,n3
  real                                  :: dz,oz,dx,ox,o3,d3
  real                                  :: aux
  real, allocatable, dimension(:,:,:)   :: modelin,modelout

  call sep_init()

  ! Model parameters
  ! Dimensions
  call from_history("n1",nzin)                  !# of samples in depth
  call from_history("o1",oz,0.0)                !origin for depth coord
  call from_history("d1",dz)                    !grid spacing in depth
  call from_history("n2",nxin)                    !# of samples in x
  call from_history("o2",ox,0.0)                !origin for x coord
  call from_history("d2",dx)                    !grid spacing in x
  call from_history("n3",n3)                    !# of samples in x
  call from_history("o3",o3,0.0)                !origin for x coord
  call from_history("d3",d3)                    !grid spacing in x

  call from_param("top",top,.true.)             !if false, there is no abc on top (free surface case)
  call from_param("abc",abc,0)                  !number os samples

  nzout = nzin + 2*abc
  if (.not.top) nzout = nzin + abc + 4 !These 4 samples are due to the derivative operator length 
  nxout = nxin + 2*abc

  allocate(modelin(nzin,nxin,n3))
  allocate(modelout(nzout,nxout,n3))

  call sep_read(modelin)

  if (top) then
    !CENTER
    modelout(abc+1:abc+nzin,abc+1:nxin+abc,:) = modelin(:,:,:)
    do ix=1,abc
      !LEFT
      modelout(abc+1:nzin+abc,ix,:) = modelin(:,1,:)
      !RIGHT
      modelout(abc+1:nzin+abc,abc+nxin+ix,:) = modelin(:,nxin,:)
    enddo
    do iz=1,abc
      !TOP
      modelout(iz,:,:) = modelout(abc+1,:,:)
      !BOTTOM
      modelout(abc+nzin+iz,:,:) = modelout(nzin+abc,:,:)
    enddo
  else
    !CENTER
    modelout(5:nzin+4,abc+1:nxin+abc,:) = modelin(:,:,:)
    do ix=1,abc
      !LEFT
      modelout(5:nzin+4,ix,:) = modelin(:,1,:)
      !RIGHT
      modelout(5:nzin+4,abc+nxin+ix,:) = modelin(:,nxin,:)
    enddo
    do iz=1,abc
      !BOTTOM
      modelout(nzin+4+iz,:,:) = modelout(nzin+4,:,:)
    enddo
    do iz=1,4
      !TOP
      modelout(iz,:,:) = modelout(5,:,:)
    enddo
  endif

  ! Write SEP header parameter
  call to_history("n1",nzout)
  call to_history("o1",oz)
  call to_history("d1",dz)
  call to_history("n2",nxout)
  call to_history("o2",ox)
  call to_history("d2",dx)
  call to_history("n3",n3)
  call to_history("o3",o3)
  call to_history("d3",d3)
  call to_history("label1","z")
  call to_history("label2","x")
  call to_history("label3","parameter")
  call sep_write(modelout)

  call sep_close

end program
