/*This is a C library for the imaging condition used in the elastic RTM code
 * 
 * Gustavo C. Alves - March, 8th, 2016 - Stanford, CA */
export void image_rho_lame_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform float s1[], uniform float s2[], uniform float dm1_x[], uniform float dm1_z[], uniform float m1_x[], uniform float m1_z[], uniform float vx[], uniform float vz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm1_x[iz+i]=0.f;
          dm1_z[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm1_x[iz+i] += dt*m1_x[iz+i]*vx[iz+i]*s1[iz+i];
        dm1_z[iz+i] += dt*m1_z[iz+i]*vz[iz+i]*s2[iz+i];
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          vx[iz+i]=0.f;
          vz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    vx[iz+i] += dt*m1_x[iz+i]*dm1_x[iz+i]*s1[iz+i];
	    vz[iz+i] += dt*m1_z[iz+i]*dm1_z[iz+i]*s2[iz+i];	  					
	  }
    }
  }
}

export void image_lambda_lame_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform uniform float s3[], uniform float s4[], uniform float dm2[], uniform float sigmaxx[], uniform float sigmazz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm2  [iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm2  [iz+i] += -dt*(sigmaxx[iz+i]+sigmazz[iz+i])*(s3[iz+i]+s4[iz+i]);
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    sigmaxx[iz+i] += -dt*dm2[iz+i]*(s3[iz+i]+s4[iz+i]);
	    sigmazz[iz+i] += -dt*dm2[iz+i]*(s3[iz+i]+s4[iz+i]);	  					
	  }
    }
  }
}

export void image_mu_lame_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform float s3[], uniform float s4[], uniform float s5[], uniform float dm3[], uniform float sigmaxx[], uniform float sigmazz[], uniform float sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm3  [iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm3  [iz+i] += -dt*(2.f*(sigmaxx[iz+i]*s3[iz+i]+sigmazz[iz+i]*s4[iz+i])
      				       	 +sigmaxz[iz+i]*s5[iz+i]);
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    sigmaxx[iz+i] += -2.f*dt*dm3[iz+i]*s3[iz+i];
	    sigmazz[iz+i] += -2.f*dt*dm3[iz+i]*s4[iz+i];
	    sigmaxz[iz+i] += -1.f*dt*dm3[iz+i]*s5[iz+i];	  					
	  }
    }
  }
}

export void image_rho_vel_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform float s1[], uniform float s2[], uniform float s3[], uniform float s4[], uniform float s5[], uniform float dm1[], uniform float dm1_x[], uniform float dm1_z[], uniform float m1_x[], uniform float m1_z[], uniform float m2[], uniform float m3[], uniform float vx[], uniform float vz[], uniform float sigmaxx[], uniform float sigmazz[], uniform float sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm1  [iz+i]=0.f;
          dm1_x[iz+i]=0.f;
          dm1_z[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm1_x[iz+i] += dt*m1_x[iz+i]*vx[iz+i]*s1[iz+i];
        dm1_z[iz+i] += dt*m1_z[iz+i]*vz[iz+i]*s2[iz+i];
        dm1  [iz+i] += dt*(-sigmaxx[iz+i]*(m2[iz+i]*m2[iz+i]*s3[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s4[iz+i])
        			       -sigmazz[iz+i]*(m2[iz+i]*m2[iz+i]*s4[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s3[iz+i])
        			       -sigmaxz[iz+i]*(m3[iz+i]*m3[iz+i]*s5[iz+i]));
      }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          vx[iz+i]=0.f;
          vz[iz+i]=0.f;
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    vx[iz+i] += dt*m1_x[iz+i]*dm1_x[iz+i]*s1[iz+i];
	    vz[iz+i] += dt*m1_z[iz+i]*dm1_z[iz+i]*s2[iz+i];	
	    sigmaxx[iz+i] += -dt*dm1[iz+i]*(m2[iz+i]*m2[iz+i]*s3[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s4[iz+i]);
	    sigmazz[iz+i] += -dt*dm1[iz+i]*(m2[iz+i]*m2[iz+i]*s4[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s3[iz+i]);
	    sigmaxz[iz+i] += -dt*dm1[iz+i]*(m3[iz+i]*m3[iz+i]*s5[iz+i]);
	  }					
    }
  }
}

export void image_vp_vel_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform float s3[], uniform float s4[], uniform float dm2[], uniform float m1[], uniform float m2[], uniform float sigmaxx[], uniform float sigmazz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm2[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm2[iz+i] += -2.f*dt*m1[iz+i]*m2[iz+i]*(s3[iz+i]+s4[iz+i])*(sigmaxx[iz+i]+sigmazz[iz+i]);
      }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {	
	    sigmaxx[iz+i] += -dt*dm2[iz+i]*2.f*m1[iz+i]*m2[iz+i]*(s3[iz+i]+s4[iz+i]);
	    sigmazz[iz+i] += -dt*dm2[iz+i]*2.f*m1[iz+i]*m2[iz+i]*(s3[iz+i]+s4[iz+i]);
	  }					
    }
  }
}

export void image_vs_vel_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform float s3[], uniform float s4[], uniform float s5[], uniform float dm3[], uniform float m1[], uniform float m3[], uniform float sigmaxx[], uniform float sigmazz[], uniform float sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm3[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm3[iz+i] += dt*m1[iz+i]*m3[iz+i]*(4.f*(s4[iz+i]*sigmaxx[iz+i]+s3[iz+i]*sigmazz[iz+i])-2.f*s5[iz+i]*sigmaxz[iz+i]);
      }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    sigmaxx[iz+i] +=  dt*dm3[iz+i]*m1[iz+i]*m3[iz+i]*4.f*s4[iz+i];
	    sigmazz[iz+i] +=  dt*dm3[iz+i]*m1[iz+i]*m3[iz+i]*4.f*s3[iz+i];
	    sigmaxz[iz+i] += -dt*dm3[iz+i]*m1[iz+i]*m3[iz+i]*2.f*s5[iz+i];
	  }					
    }
  }
}

export void lintx_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz,  uniform float model[], uniform float data[]){

  if (adj==1) {
    if (add==0) {
      foreach(iz=0...nz*nx) {
        model[iz]=0.f;
      }
    }
    for (uniform int ix=1;ix<nx;ix++) {
      uniform int i=(ix-1)*nz;
      foreach(iz=0...nz) {
        model[iz+i]    += 0.5*data[iz+i];
        model[iz+i+nz] += 0.5*data[iz+i];
	  }
	}
    foreach(iz=0...nz) {
	  model[iz+(nx-1)*nz] += data[iz+(nx-1)*nz];
	}
  }
  else {
    if (add==0) {
      foreach(iz=0...nz*nx) {
          data[iz]=0.f;
      }
    }
    for (uniform int ix=1;ix<nx;ix++) {
      uniform int i=(ix-1)*nz;
      foreach(iz=0...nz) {
        data[iz+i] += 0.5*(model[iz+i]+model[iz+i+nz]);
	  }
	}
	foreach(iz=0...nz) {
	  data[iz+(nx-1)*nz] += model[iz+(nx-1)*nz];
	}
  }
}

export void lintz_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz,  uniform float model[], uniform float data[]){

  if (adj==1) {
    for (uniform int ix=1;ix<=nx;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          model[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz-1) {
        model[iz+i] += 0.5*data[iz+i];
        model[iz+i+1] += 0.5*data[iz+i];
	  }
	  model[(nz-1)+i] += data[(nz-1)+i];
	}
  }
  else {
    for (uniform int ix=1;ix<=nx;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          data[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz-1) {
        data[iz+i] += 0.5*(model[iz+i]+model[iz+i+1]);
	  }
	  data[(nz-1)+i] += model[(nz-1)+i];
	}
  }
}

export void lintxz_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz,  uniform float model[], uniform float data[]){

  if (adj==1) {
    for (uniform int ix=1;ix<nx;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          model[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz-1) {
        model[iz+i  ]    += 0.25*data[iz+i];
        model[iz+i+1]    += 0.25*data[iz+i];
        model[iz+i+nz]   += 0.25*data[iz+i];
        model[iz+i+nz+1] += 0.25*data[iz+i];   
	  }
	  model[(nz-1)+i] += data[(nz-1)+i];
	}
	foreach(iz=0...nz) {
	  model[iz+(nx-1)*nz] += data[iz+(nx-1)*nz];
	}
  }
  else {
    for (uniform int ix=1;ix<nx;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          data[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz-1) {
        data[iz+i] += 0.25*(model[iz+i]+model[iz+i+1]+model[iz+i+nz]+model[iz+i+nz+1]);     
	  }
	  data[(nz-1)+i] += model[(nz-1)+i];
	}
	foreach(iz=0...nz) {
	  data[iz+(nx-1)*nz] += model[iz+(nx-1)*nz];
	}	
  }
}

//////////////////
// DOUBLE PREC. 
//////////////////

export void image_rho_lame_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s1[], uniform double s2[], uniform double dm1_x[], uniform double dm1_z[], uniform double m1_x[], uniform double m1_z[], uniform double vx[], uniform double vz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm1_x[iz+i]=0.f;
          dm1_z[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm1_x[iz+i] += dt*m1_x[iz+i]*vx[iz+i]*s1[iz+i];
        dm1_z[iz+i] += dt*m1_z[iz+i]*vz[iz+i]*s2[iz+i];
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          vx[iz+i]=0.f;
          vz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    vx[iz+i] += dt*m1_x[iz+i]*dm1_x[iz+i]*s1[iz+i];
	    vz[iz+i] += dt*m1_z[iz+i]*dm1_z[iz+i]*s2[iz+i];	  					
	  }
    }
  }
}

export void image_lambda_lame_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s3[], uniform double s4[], uniform double dm2[], uniform double sigmaxx[], uniform double sigmazz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm2  [iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm2  [iz+i] += -dt*(sigmaxx[iz+i]+sigmazz[iz+i])*(s3[iz+i]+s4[iz+i]);
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    sigmaxx[iz+i] += -dt*dm2[iz+i]*(s3[iz+i]+s4[iz+i]);
	    sigmazz[iz+i] += -dt*dm2[iz+i]*(s3[iz+i]+s4[iz+i]);	  					
	  }
    }
  }
}

export void image_mu_lame_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s3[], uniform double s4[], uniform double s5[], uniform double dm3[], uniform double sigmaxx[], uniform double sigmazz[], uniform double sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm3  [iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm3  [iz+i] += -dt*(2.f*(sigmaxx[iz+i]*s3[iz+i]+sigmazz[iz+i]*s4[iz+i])
      				       	 +sigmaxz[iz+i]*s5[iz+i]);
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    sigmaxx[iz+i] += -2.f*dt*dm3[iz+i]*s3[iz+i];
	    sigmazz[iz+i] += -2.f*dt*dm3[iz+i]*s4[iz+i];
	    sigmaxz[iz+i] += -1.f*dt*dm3[iz+i]*s5[iz+i];	  					
	  }
    }
  }
}

export void image_rho_vel_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s1[], uniform double s2[], uniform double s3[], uniform double s4[], uniform double s5[], uniform double dm1[], uniform double dm1_x[], uniform double dm1_z[], uniform double m1_x[], uniform double m1_z[], uniform double m2[], uniform double m3[], uniform double vx[], uniform double vz[], uniform double sigmaxx[], uniform double sigmazz[], uniform double sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm1  [iz+i]=0.f;
          dm1_x[iz+i]=0.f;
          dm1_z[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm1_x[iz+i] += dt*m1_x[iz+i]*vx[iz+i]*s1[iz+i];
        dm1_z[iz+i] += dt*m1_z[iz+i]*vz[iz+i]*s2[iz+i];
        dm1  [iz+i] += dt*(-sigmaxx[iz+i]*(m2[iz+i]*m2[iz+i]*s3[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s4[iz+i])
        			       -sigmazz[iz+i]*(m2[iz+i]*m2[iz+i]*s4[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s3[iz+i])
        			       -sigmaxz[iz+i]*(m3[iz+i]*m3[iz+i]*s5[iz+i]));
      }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          vx[iz+i]=0.f;
          vz[iz+i]=0.f;
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    vx[iz+i] += dt*m1_x[iz+i]*dm1_x[iz+i]*s1[iz+i];
	    vz[iz+i] += dt*m1_z[iz+i]*dm1_z[iz+i]*s2[iz+i];	
	    sigmaxx[iz+i] += -dt*dm1[iz+i]*(m2[iz+i]*m2[iz+i]*s3[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s4[iz+i]);
	    sigmazz[iz+i] += -dt*dm1[iz+i]*(m2[iz+i]*m2[iz+i]*s4[iz+i]+(m2[iz+i]*m2[iz+i]-2.f*m3[iz+i]*m3[iz+i])*s3[iz+i]);
	    sigmaxz[iz+i] += -dt*dm1[iz+i]*(m3[iz+i]*m3[iz+i]*s5[iz+i]);
	  }					
    }
  }
}

export void image_vp_vel_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s3[], uniform double s4[], uniform double dm2[], uniform double m1[], uniform double m2[], uniform double sigmaxx[], uniform double sigmazz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm2[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm2[iz+i] += -2.f*dt*m1[iz+i]*m2[iz+i]*(s3[iz+i]+s4[iz+i])*(sigmaxx[iz+i]+sigmazz[iz+i]);
      }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {	
	    sigmaxx[iz+i] += -dt*dm2[iz+i]*2.f*m1[iz+i]*m2[iz+i]*(s3[iz+i]+s4[iz+i]);
	    sigmazz[iz+i] += -dt*dm2[iz+i]*2.f*m1[iz+i]*m2[iz+i]*(s3[iz+i]+s4[iz+i]);
	  }					
    }
  }
}

export void image_vs_vel_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s3[], uniform double s4[], uniform double s5[], uniform double dm3[], uniform double m1[], uniform double m3[], uniform double sigmaxx[], uniform double sigmazz[], uniform double sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          dm3[iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm3[iz+i] += dt*m1[iz+i]*m3[iz+i]*(4.f*(s4[iz+i]*sigmaxx[iz+i]+s3[iz+i]*sigmazz[iz+i])-2.f*s5[iz+i]*sigmaxz[iz+i]);
      }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        foreach(iz=0...nz) {
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    sigmaxx[iz+i] +=  dt*dm3[iz+i]*m1[iz+i]*m3[iz+i]*4.f*s4[iz+i];
	    sigmazz[iz+i] +=  dt*dm3[iz+i]*m1[iz+i]*m3[iz+i]*4.f*s3[iz+i];
	    sigmaxz[iz+i] += -dt*dm3[iz+i]*m1[iz+i]*m3[iz+i]*2.f*s5[iz+i];
	  }					
    }
  }
}

export void image_lame_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform float dt, uniform int bx, uniform int ex, uniform float s1[], uniform float s2[], uniform float s3[], uniform float s4[], uniform float s5[], uniform float dm1_x[], uniform float dm1_z[], uniform float dm2[], uniform float dm3[], uniform float m1_x[], uniform float m1_z[], uniform float vx[], uniform float vz[], uniform float sigmaxx[], uniform float sigmazz[], uniform float sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        for (uniform int iz=0;iz<nz;iz++) {
          dm1_x[iz+i]=0.f;
          dm1_z[iz+i]=0.f;
          dm2  [iz+i]=0.f;
          dm3  [iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm1_x[iz+i] += dt*m1_x[iz+i]*vx[iz+i]*s1[iz+i];
        dm1_z[iz+i] += dt*m1_z[iz+i]*vz[iz+i]*s2[iz+i];
        dm2  [iz+i] += -dt*(sigmaxx[iz+i]+sigmazz[iz+i])*(s3[iz+i]+s4[iz+i]);
        dm3  [iz+i] += -dt*(2.f*(sigmaxx[iz+i]*s3[iz+i]+sigmazz[iz+i]*s4[iz+i])
      					     +sigmaxz[iz+i]*s5[iz+i]);
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        for (uniform int iz=0;iz<nz;iz++) {
          vx[iz+i]=0.f;
          vz[iz+i]=0.f;
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    vx[iz+i] += dt*m1_x[iz+i]*dm1_x[iz+i]*s1[iz+i];
	    vz[iz+i] += dt*m1_z[iz+i]*dm1_z[iz+i]*s2[iz+i];
	    sigmaxx[iz+i] += -dt*(dm2[iz+i]*(s3[iz+i]+s4[iz+i])
	  					     +dm3[iz+i]*2.f*s3[iz+i]);
	    sigmazz[iz+i] += -dt*(dm2[iz+i]*(s3[iz+i]+s4[iz+i])
	  					     +dm3[iz+i]*2.f*s4[iz+i]);
	    sigmaxz[iz+i] += -dt*(dm3[iz+i]*s5[iz+i]);	  					
	  }
    }
  }
}

export void image_lame_double_c_(uniform int adj, uniform int add, uniform int nx, uniform int nz, uniform double dt, uniform int bx, uniform int ex, uniform double s1[], uniform double s2[], uniform double s3[], uniform double s4[], uniform double s5[],  uniform double dm1_x[], uniform double dm1_z[], uniform double dm2[], uniform double dm3[], uniform double m1_x[], uniform double m1_z[], uniform double vx[], uniform double vz[], uniform double sigmaxx[], uniform double sigmazz[], uniform double sigmaxz[]){

  if (adj==1) {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        for (uniform int iz=0;iz<nz;iz++) {
          dm1_x[iz+i]=0.f;
          dm1_z[iz+i]=0.f;
          dm2  [iz+i]=0.f;
          dm3  [iz+i]=0.f;
        }
      }
      foreach(iz=0...nz) {
        dm1_x[iz+i] += dt*m1_x[iz+i]*vx[iz+i]*s1[iz+i];
        dm1_z[iz+i] += dt*m1_z[iz+i]*vz[iz+i]*s2[iz+i];
        dm2  [iz+i] += -dt*(sigmaxx[iz+i]+sigmazz[iz+i])*(s3[iz+i]+s4[iz+i]);
        dm3  [iz+i] += -dt*(2.f*(sigmaxx[iz+i]*s3[iz+i]+sigmazz[iz+i]*s4[iz+i])
      					     +sigmaxz[iz+i]*s5[iz+i]);
	  }
    }
  }
  else {
    for (uniform int ix=bx;ix<ex;ix++) {
      uniform int i=(ix-1)*nz;
      if (add==0) {
        for (uniform int iz=0;iz<nz;iz++) {
          vx[iz+i]=0.f;
          vz[iz+i]=0.f;
          sigmaxx[iz+i]=0.f;
          sigmazz[iz+i]=0.f;
          sigmaxz[iz+i]=0.f;
        }
      }
	  foreach(iz=0...nz) {
	    vx[iz+i] += dt*m1_x[iz+i]*dm1_x[iz+i]*s1[iz+i];
	    vz[iz+i] += dt*m1_z[iz+i]*dm1_z[iz+i]*s2[iz+i];
	    sigmaxx[iz+i] += -dt*(dm2[iz+i]*(s3[iz+i]+s4[iz+i])
	  					     +dm3[iz+i]*2.f*s3[iz+i]);
	    sigmazz[iz+i] += -dt*(dm2[iz+i]*(s3[iz+i]+s4[iz+i])
	  					     +dm3[iz+i]*2.f*s4[iz+i]);
	    sigmaxz[iz+i] += -dt*(dm3[iz+i]*s5[iz+i]);	  					
	  }
    }
  }
}