/*This is a C library for the linear interpolations used in the elastic RTM code
 * 
 * Gustavo C. Alves - September, 10th, 2015 - Stanford, CA */

export void lint_serial_c_(uniform int adj, uniform int add, uniform int n, uniform float dt, uniform float dtseis, uniform float model[], uniform float data[], uniform float aux, uniform int it){

  uniform float temp=(it*dt)/dtseis;
  uniform int   itseis=temp;
  uniform float f=temp-itseis;
  uniform float g=1.0f-f;

  if (adj==1) {
    if (add==0) {
      foreach (i=0...n){
          model[i] = 0.f;
	  }
    }
    foreach(i=0...n) {
      model[i] += aux*g*data[i];
      model[i+n] += aux*f*data[i];
    }
  }
  else {
    if (add==0) {
      foreach (i=0...n){
        data[i] = 0.f;
	  }
	}
    foreach(i=0...n) {
      data[i] += aux*(g*model[i]+f*model[i+n]);
    }
  }
}

export void lint_serial2_c_(uniform int adj, uniform int add, uniform int n, uniform float dt, uniform float dz, uniform float dtseis, uniform float model[], uniform float data[], uniform float aux, uniform int it,uniform float recz){

  uniform float tempt=(it*dt)/dtseis;
  uniform int   itseis=tempt;
  uniform float ft=tempt-itseis;
  uniform float gt=1.0f-ft;
  uniform float tempz=recz/dz;
  uniform int   iz=tempz;
  uniform float fz=tempz-iz;
  uniform float gz=1.0f-fz;

  if (adj==1) {
    if (add==0) {
      foreach (i=0...n){
          model[i] = 0.f;
		  model[i+n] = 0.f;
	  }
    }
    foreach(i=0...n) {
      model[i] += aux*gt*(gz*data[2*i]+fz*data[2*i+1]);
      model[i+n] += aux*ft*(gz*data[2*i]+fz*data[2*i+1]);
    }
  }
  else {
    if (add==0) {
      foreach (i=0...n){
        data[i] = 0.f;
        data[i+n] = 0.f;
	  }
	}
    foreach(i=0...n) {
      data[2*i] += aux*gz*(gt*model[i]+ft*model[i+n]);
      data[2*i+1] += aux*fz*(gt*model[i]+ft*model[i+n]);
    }
  }
}

export void lint_par_c_(uniform int adj, uniform int add, uniform int nz, uniform int nx, uniform int ix, uniform float dt, uniform float dtseis, uniform float model[], uniform float data[], uniform float aux, uniform int it){

  uniform float temp=(it*dt)/dtseis;
  uniform int   itseis=temp;
  uniform float f=temp-itseis;
  uniform float g=1.0f-f;
  uniform int   i=(ix-1)*nz;
  uniform int   n=nx*nz;

  if (adj==1) {
    if (add==0) {
      foreach(iz=0...nz) {
        model[iz+i]=0.f;
        model[iz+i+n]=0.f;
      }
    }
    foreach(iz=0...nz) {
      model[iz+i]   += aux*g*data[iz+i];
      model[iz+i+n] += aux*f*data[iz+i];
    }
  }
  else {
    if (add==0) {
      foreach(iz=0...nz) {
        data[iz+i]=0.f;
      }
    }
    foreach(iz=0...nz) {
      data[iz+i] += aux*(g*model[iz+i]+f*model[iz+i+n]);
    }
  }
}

export void lint_par2_c_(uniform int adj, uniform int add, uniform int nz, uniform float dt, uniform float dtseis, uniform float model1[], uniform float model2[], uniform float data[], uniform float aux, uniform int it, uniform int ix){

  uniform float temp=(it*dt)/dtseis;
  uniform int   itseis=temp;
  uniform float f=temp-itseis;
  uniform float g=1.0f-f;
  uniform int   i=(ix-1)*nz;

  if (adj==1) {
    if (add==0) {
      foreach(iz=0...nz) {
        model1[iz+i]=0.f;
        model2[iz+i]=0.f;
      }
    }      
    foreach(iz=0...nz) {
      model1[iz+i] += aux*g*data[iz+i];
      model2[iz+i] += aux*f*data[iz+i];
    }
  }
  else {
    if (add==0) {
      foreach(iz=0...nz) {
        data[iz+i]=0.f;
      }
    }
    foreach(iz=0...nz) {
      data[iz+i] += aux*(g*model1[iz+i]+f*model2[iz+i]);
    }
  }
}

export void lint_serial_double_c_(uniform int adj, uniform int add, uniform int n, uniform float dt, uniform float dtseis, uniform double model[], uniform double data[], uniform double aux, uniform int it){

  uniform float temp=(it*dt)/dtseis;
  uniform int   itseis=temp;
  uniform float f=temp-itseis;
  uniform float g=1.0f-f;

  if (adj==1) {
    if (add==0) {
      foreach (i=0...n){
          model[i] = 0.f;
          model[i+n] = 0.f;
	  }
    }
    foreach(i=0...n) {
      model[i] += aux*g*data[i];
      model[i+n] += aux*f*data[i];
    }
  }
  else {
    if (add==0) {
      foreach (i=0...n){
        data[i] = 0.f;
	  }
	}
    foreach(i=0...n) {
      data[i] += aux*(g*model[i]+f*model[i+n]);
    }
  }
}

export void lint_serial2_double_c_(uniform int adj, uniform int add, uniform int n, uniform float dt, uniform float dz, uniform float dtseis, uniform double model[], uniform double data[], uniform double aux, uniform int it,uniform float recz){

  uniform float tempt=(it*dt)/dtseis;
  uniform int   itseis=tempt;
  uniform float ft=tempt-itseis;
  uniform float gt=1.0f-ft;
  uniform float tempz=recz/dz;
  uniform int   iz=tempz;
  uniform float fz=tempz-iz;
  uniform float gz=1.0f-fz;

  if (adj==1) {
    if (add==0) {
      foreach (i=0...n){
          model[i] = 0.f;
		  model[i+n] = 0.f;
	  }
    }
    foreach(i=0...n) {
      model[i] += aux*gt*(gz*data[2*i]+fz*data[2*i+1]);
      model[i+n] += aux*ft*(gz*data[2*i]+fz*data[2*i+1]);
    }
  }
  else {
    if (add==0) {
      foreach (i=0...n){
        data[i] = 0.f;
        data[i+n] = 0.f;
	  }
	}
    foreach(i=0...n) {
      data[2*i] += aux*gz*(gt*model[i]+ft*model[i+n]);
      data[2*i+1] += aux*fz*(gt*model[i]+ft*model[i+n]);
    }
  }
}

export void lint_par_double_c_(uniform int adj, uniform int add, uniform int nz, uniform int nx, uniform int ix, uniform float dt, uniform float dtseis, uniform double model[], uniform double data[], uniform double aux, uniform int it){

  uniform float temp=(it*dt)/dtseis;
  uniform int   itseis=temp;
  uniform float f=temp-itseis;
  uniform float g=1.0f-f;
  uniform int   i=(ix-1)*nz;
  uniform int   n=nx*nz;

  if (adj==1) {
    if (add==0) {
      foreach(iz=0...nz) {
        model[iz+i]=0.f;
        model[iz+i+n]=0.f;
      }
    }
    foreach(iz=0...nz) {
      model[iz+i]   += aux*g*data[iz+i];
      model[iz+i+n] += aux*f*data[iz+i];
    }
  }
  else {
    if (add==0) {
      foreach(iz=0...nz) {
        data[iz+i]=0.f;
      }
    }
    foreach(iz=0...nz) {
      data[iz+i] += aux*(g*model[iz+i]+f*model[iz+i+n]);
    }
  }
}

export void lint_par2_double_c_(uniform int adj, uniform int add, uniform int nz, uniform float dt, uniform float dtseis, uniform double model1[], uniform double model2[], uniform double data[], uniform double aux, uniform int it, uniform int ix){

  uniform float temp=(it*dt)/dtseis;
  uniform int   itseis=temp;
  uniform float f=temp-itseis;
  uniform float g=1.0f-f;
  uniform int   i=(ix-1)*nz;

  if (adj==1) {
    if (add==0) {
      foreach(iz=0...nz) {
        model1[iz+i]=0.f;
        model2[iz+i]=0.f;
      }
    }      
    foreach(iz=0...nz) {
      model1[iz+i] += aux*g*data[iz+i];
      model2[iz+i] += aux*f*data[iz+i];
    }
  }
  else {
    if (add==0) {
      foreach(iz=0...nz) {
        data[iz+i]=0.f;
      }
    }
    foreach(iz=0...nz) {
      data[iz+i] += aux*(g*model1[iz+i]+f*model2[iz+i]);
    }
  }
}

