/*This is a C library for the sinc interpolations used in the elastic RTM code
 * 
 * Gustavo C. Alves - November, 10th, 2016 - Stanford, CA */

export uniform float sinc_adj_serial_c_(uniform int add, uniform int nf, uniform int ixrec, uniform int izrec, uniform float xrec, uniform float zrec, uniform float filter[], uniform float data[]){

  uniform float fx=xrec-ixrec;
  uniform float gx=1.0f-fx;
  uniform float fz=zrec-izrec;
  uniform float gz=1.0f-fz;
  uniform float model;

  if (add==0) model = 0.f;
  for (uniform int ix=0;ix<nf;ix++) {
    uniform int i=ix*nf;
	for (uniform int iz=0;iz<nf;iz++) {
      model += filter[iz+i] * ( gx * gz * data[iz+i     ] +
      							gx * fz * data[iz+i+1   ] +
      							fx * gz * data[iz+i  +nf] +
      							fx * fz * data[iz+i+1+nf]);
    }
  }
  return model;
}

export void sinc_serial_c_(uniform int add, uniform int nf, uniform int ixrec, uniform int izrec, uniform float xrec, uniform float zrec, uniform float model, uniform float filter[], uniform float data[]){

  uniform float fx=xrec-ixrec;
  uniform float gx=1.0f-fx;
  uniform float fz=zrec-izrec;
  uniform float gz=1.0f-fz;
  uniform int   n2=(nf+1)*(nf+1);

  if (add==0) {
    for (uniform int i=0;i<n2;i++) {
      data[i] = 0.f;
	}
  }
  for (uniform int ix=0;ix<nf;ix++) {
    uniform int i=ix*nf;
	for (uniform int iz=0;iz<nf;iz++) {
      data[iz+i     ] += filter[iz+i] * gx * gz * model;
      data[iz+i+1   ] += filter[iz+i] * gx * fz * model;
      data[iz+i  +nf] += filter[iz+i] * fx * gz * model;
      data[iz+i+1+nf] += filter[iz+i] * fx * fz * model;
    }
  }
}

export void sinc_serial2_c_(uniform int adj, uniform int add, uniform int nf, uniform int nz, uniform int nx, uniform float xrec, uniform float zrec, uniform float dx, uniform int it, uniform float dt, uniform float dtseis, uniform int ntseis, uniform int irec, uniform int nrec, uniform float aux, uniform float model[], uniform float filter[], uniform float data[]){

  uniform float tempt=(it*dt)/dtseis;
  uniform int   itseis=tempt;
  uniform float ft=tempt-itseis;
  uniform float gt=1.0f-ft;
  uniform int   it2=(itseis*nrec)+irec-1;
  uniform float tempx=xrec/dx;
  uniform int   itempx=tempx;
  uniform float fx=tempx-itempx;
  uniform float gx=1.0f-fx;
  uniform float tempz=zrec/dx;
  uniform int   itempz=tempz;
  uniform float fz=tempz-itempz;
  uniform float gz=1.0f-fz;
  uniform int   nm=(nf-1)/2;
  uniform int   xbeg=itempx-nm-1;
  uniform int   zbeg=itempz-nm-1;

  if (adj==1) {
    if (add==0) {
      for (uniform int i=0;i<nz*nx;i++) {
        model[i] = 0.f;
	  }
    }
    for (uniform int ix=0;ix<nf;ix++) {
      uniform int i1=ix*nf;
      uniform int i2=(ix+xbeg)*nz;
	  for (uniform int iz=0;iz<nf;iz++) {
        model[iz+i2     +zbeg] += aux*filter[iz+i1]*gx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+1   +zbeg] += aux*filter[iz+i1]*gx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2  +nz+zbeg] += aux*filter[iz+i1]*fx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+1+nz+zbeg] += aux*filter[iz+i1]*fx*fz*(gt*data[it2]+ft*data[it2+nrec]);
      }     							  
    }
  } 
  else {
    if (add==0) {
      for (uniform int i=0;i<nrec*ntseis;i++) {
        data[i] = 0.f;
	  }
    }
    for (uniform int ix=0;ix<nf;ix++) {
      uniform int i1=ix*nf;
      uniform int i2=(ix+xbeg)*nz;
	  for (uniform int iz=0;iz<nf;iz++) {
        data[it2] += aux*gt*filter[iz+i1]*(gx*gz*model[iz+i2     +zbeg] +
      							  		   gx*fz*model[iz+i2+1   +zbeg] +
      							  		   fx*gz*model[iz+i2  +nz+zbeg] +
      							  		   fx*fz*model[iz+i2+1+nz+zbeg]);
        data[it2+nrec] += aux*ft*filter[iz+i1]*(gx*gz*model[iz+i2     +zbeg] +
      							  		        gx*fz*model[iz+i2+1   +zbeg] +
      							  		        fx*gz*model[iz+i2  +nz+zbeg] +
      							  		        fx*fz*model[iz+i2+1+nz+zbeg]);
      }
    }
  }
}

export void sinc_serial2_fz_c_(uniform int adj, uniform int add, uniform int nf, uniform int nz, uniform int nx, uniform float xrec, uniform float zrec, uniform float dx, uniform int it, uniform float dt, uniform float dtseis, uniform int ntseis, uniform int irec, uniform int nrec, uniform float aux, uniform float model[], uniform float filter[], uniform float data[]){

  uniform float tempt=(it*dt)/dtseis;
  uniform int   itseis=tempt;
  uniform float ft=tempt-itseis;
  uniform float gt=1.0f-ft;
  uniform int   it2=(itseis*nrec)+irec-1;
  uniform float tempx=xrec/dx;
  uniform int   itempx=tempx;
  uniform float fx=tempx-itempx;
  uniform float gx=1.0f-fx;
  uniform float tempz=zrec/dx;
  uniform int   itempz=tempz;
  uniform float fz=tempz-itempz;
  uniform float gz=1.0f-fz;
  uniform int   nm=(nf-1)/2;
  uniform int   xbeg=itempx-nm-1;
  uniform int   zbeg=itempz-nm-1;

  if (adj==1) {
    if (add==0) {
      for (uniform int i=0;i<nz*nx;i++) {
        model[i] = 0.f;
	  }
    }
    for (uniform int ix=0;ix<nf;ix++) {
      uniform int i1=ix*nf;
      uniform int i2=(ix+xbeg)*nz;
	  for (uniform int iz=0;iz<nf;iz++) {
        model[iz+i2-1   +zbeg] +=-aux*filter[iz+i1]*gx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2     +zbeg] +=-aux*filter[iz+i1]*gx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2-1+nz+zbeg] +=-aux*filter[iz+i1]*fx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2  +nz+zbeg] +=-aux*filter[iz+i1]*fx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+1   +zbeg] += aux*filter[iz+i1]*gx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+2   +zbeg] += aux*filter[iz+i1]*gx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+1+nz+zbeg] += aux*filter[iz+i1]*fx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+2+nz+zbeg] += aux*filter[iz+i1]*fx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        
      }     							  
    }
  } 
  else {
    if (add==0) {
      for (uniform int i=0;i<nrec*ntseis;i++) {
        data[i] = 0.f;
	  }
    }
    for (uniform int ix=0;ix<nf;ix++) {
      uniform int i1=ix*nf;
      uniform int i2=(ix+xbeg)*nz;
	  for (uniform int iz=0;iz<nf;iz++) {
        data[it2] += aux*gt*filter[iz+i1]*(-(gx*gz*model[iz+i2-1   +zbeg] +
      							  		     gx*fz*model[iz+i2     +zbeg] +
      							  		     fx*gz*model[iz+i2-1+nz+zbeg] +
      							  		     fx*fz*model[iz+i2  +nz+zbeg])
      							  		   +(gx*gz*model[iz+i2+1   +zbeg] +
      							  		     gx*fz*model[iz+i2+2   +zbeg] +
      							  		     fx*gz*model[iz+i2+1+nz+zbeg] +
      							  		     fx*fz*model[iz+i2+2+nz+zbeg]));
        data[it2+nrec] += aux*ft*filter[iz+i1]*(-(gx*gz*model[iz+i2-1   +zbeg] +
      							  		     	  gx*fz*model[iz+i2     +zbeg] +
      							  		     	  fx*gz*model[iz+i2-1+nz+zbeg] +
      							  		     	  fx*fz*model[iz+i2  +nz+zbeg])
      							  		   		+(gx*gz*model[iz+i2+1   +zbeg] +
      							  		     	  gx*fz*model[iz+i2+2   +zbeg] +
      							  		     	  fx*gz*model[iz+i2+1+nz+zbeg] +
      							  		     	  fx*fz*model[iz+i2+2+nz+zbeg]));
      }
    }
  }
}

export void sinc_serial2_fx_c_(uniform int adj, uniform int add, uniform int nf, uniform int nz, uniform int nx, uniform float xrec, uniform float zrec, uniform float dx, uniform int it, uniform float dt, uniform float dtseis, uniform int ntseis, uniform int irec, uniform int nrec, uniform float aux, uniform float model[], uniform float filter[], uniform float data[]){

  uniform float tempt=(it*dt)/dtseis;
  uniform int   itseis=tempt;
  uniform float ft=tempt-itseis;
  uniform float gt=1.0f-ft;
  uniform int   it2=(itseis*nrec)+irec-1;
  uniform float tempx=xrec/dx;
  uniform int   itempx=tempx;
  uniform float fx=tempx-itempx;
  uniform float gx=1.0f-fx;
  uniform float tempz=zrec/dx;
  uniform int   itempz=tempz;
  uniform float fz=tempz-itempz;
  uniform float gz=1.0f-fz;
  uniform int   nm=(nf-1)/2;
  uniform int   xbeg=itempx-nm-1;
  uniform int   zbeg=itempz-nm-1;

  if (adj==1) {
    if (add==0) {
      for (uniform int i=0;i<nz*nx;i++) {
        model[i] = 0.f;
	  }
    }
    for (uniform int ix=0;ix<nf;ix++) {
      uniform int i1=ix*nf;
      uniform int i2=(ix+xbeg)*nz;
	  for (uniform int iz=0;iz<nf;iz++) {
        model[iz+i2-  nz   +zbeg] +=-aux*filter[iz+i1]*gx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2        +zbeg] +=-aux*filter[iz+i1]*gx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2-  nz+nz+zbeg] +=-aux*filter[iz+i1]*fx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2     +nz+zbeg] +=-aux*filter[iz+i1]*fx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+  nz   +zbeg] += aux*filter[iz+i1]*gx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+2*nz   +zbeg] += aux*filter[iz+i1]*gx*fz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+  nz+nz+zbeg] += aux*filter[iz+i1]*fx*gz*(gt*data[it2]+ft*data[it2+nrec]);
        model[iz+i2+2*nz+nz+zbeg] += aux*filter[iz+i1]*fx*fz*(gt*data[it2]+ft*data[it2+nrec]);        
      }     							  
    }
  } 
  else {
    if (add==0) {
      for (uniform int i=0;i<nrec*ntseis;i++) {
        data[i] = 0.f;
	  }
    }
    for (uniform int ix=0;ix<nf;ix++) {
      uniform int i1=ix*nf;
      uniform int i2=(ix+xbeg)*nz;
	  for (uniform int iz=0;iz<nf;iz++) {
        data[it2] += aux*gt*filter[iz+i1]*(-(gx*gz*model[iz+i2-  nz   +zbeg] +
      							  		     gx*fz*model[iz+i2        +zbeg] +
      							  		     fx*gz*model[iz+i2-  nz+nz+zbeg] +
      							  		     fx*fz*model[iz+i2     +nz+zbeg])
      							  		   +(gx*gz*model[iz+i2+  nz   +zbeg] +
      							  		     gx*fz*model[iz+i2+2*nz   +zbeg] +
      							  		     fx*gz*model[iz+i2+  nz+nz+zbeg] +
      							  		     fx*fz*model[iz+i2+2*nz+nz+zbeg]));
        data[it2+nrec] += aux*ft*filter[iz+i1]*(-(gx*gz*model[iz+i2-  nz   +zbeg] +
      							  		     	  gx*fz*model[iz+i2        +zbeg] +
      							  		     	  fx*gz*model[iz+i2-  nz+nz+zbeg] +
      							  		     	  fx*fz*model[iz+i2     +nz+zbeg])
      							  		   		+(gx*gz*model[iz+i2+  nz   +zbeg] +
      							  		     	  gx*fz*model[iz+i2+2*nz   +zbeg] +
      							  		     	  fx*gz*model[iz+i2+  nz+nz+zbeg] +
      							  		     	  fx*fz*model[iz+i2+2*nz+nz+zbeg]));
      }
    }
  }
}