! Main program to create 2D models
! Uses a txt file with x,z pairs for the node positions as a reference for the water bottom
program create_src_rec_list

  use sep

  implicit none

  integer								:: nsrc,nrec,inrec
  real									:: dsrc,osrc,drec,orec
  real, allocatable, dimension(:,:)     :: rec_list
  real, allocatable, dimension(:)       :: src_list

  call sep_init()

  ! Model parameters
  ! Dimensions
  call from_param("nsrc",nsrc)                    !# of samples in depth
  call from_param("osrc",osrc,0.0)                !origin for depth coord
  call from_param("dsrc",dsrc)                    !grid spacing in depth
  call from_param("nrec",nrec)                    !# of samples in depth
  call from_param("orec",orec,0.0)                !origin for depth coord
  call from_param("drec",drec)                    !grid spacing in depth  

  ! Allocation of the input/output matrices
  allocate(rec_list(nrec,3))
  allocate(src_list(nsrc))

  open(unit=20,file="Dat_bkp/all-nodes.txt")

  do inrec=1,nrec
    read(20,*)rec_list(inrec,1),rec_list(inrec,2),rec_list(inrec,3)
  enddo

  close(20)

  open(unit=20,file="Dat_bkp/all-shots.txt")
  
  do inrec=1,nsrc
    read(20,*)src_list(inrec)
  enddo
  
  close(20)

  do inrec=1,nrec
    rec_list(inrec,2) = 1000.0*rec_list(inrec,2)-40000
	rec_list(inrec,3) =-1000.0*rec_list(inrec,3)
  enddo

  do inrec=1,nsrc
    src_list(inrec) = 1000.0*src_list(inrec)-40000
  enddo

  ! Write SEP header parameter
  call to_history("n1",nsrc)
  call to_history("o1",osrc)
  call to_history("d1",dsrc)
  call to_history("n2",3)
  call to_history("o2",0.)
  call to_history("d2",1)
  
  call sep_write(rec_list)
  
  call to_history("n1",nrec,"src_list")
  call to_history("o1",orec,"src_list")
  call to_history("d1",drec,"src_list")
  
  call sep_write(src_list,"src_list")

  call sep_close

end program
