! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! June, 2016

! RTM code that is supposed to take a forward/adjoint
! input, so that it can be easily tested/put into an
! iterative solver.

module elastic_mod

  use abc_mod
  use model_mod
  use clib_rtm_ispc
  use clib_lint_ispc
  use clib_sinc_ispc
  use sep
  use source_mod
  use seis_mod_rtm
  use snap_mod_rtm
  use omp_lib
  use sinc_mod

  implicit none
  integer, private :: it, ntseis, itseis, nx, nz, abc, nf, nm, nrec, izrec, source_type, irec, ixrec, dtratio,data_comp
  real, private    :: dt, dt2, dtseis, dx, dz, fp, aux, aux2, aux3, aux4, fc, xsource, zsource, zrec, xrec
  logical, private :: lame, snap, write_bkg, absorbingBound, freeSurface, rec_list,adj
  integer*8,private :: nt, long1,long2
  integer, private, parameter :: T=1,F=0

  real, pointer, dimension(:,:),private        :: m1,m2,m3
  real, allocatable, dimension(:,:), private   :: m1_x,m1_z
  real, allocatable, dimension(:,:), private   :: aux_m1_x,aux_m1_z,aux_m2,aux_m3,aux_m2m3

  !Source
  real, allocatable, dimension(:), private	:: source,source_wavelet
  !Sinc filter
  real, allocatable, dimension(:,:), private:: filter
  !Receiver list file
  real, allocatable, dimension(:), private	:: rec_coords

  !Wavefields  
  real, pointer, dimension(:),   private    :: pwaveNew,pwaveCur,pwaveTemp
  real, target, allocatable, dimension(:)	:: waveNew,waveCur

  !Data
  real, pointer, dimension(:),   private    :: prec
  
  !Nodes thread split
  integer, private							  :: nodes
  integer, pointer, dimension(:), private     :: b,e

  !TEMP
  !real, allocatable, dimension(:), private, target :: tempCur1
   
  contains

  ! Initialization of all variables needed by this subroutine
  subroutine elastic_init(adj_in,nx_in,dx_in,nz_in,dz_in,dtseis_in,ntseis_in,rec_length_in,rec_list_in,zrec_in,xrec_in,xsource_in,zsource_in,source_type_in,data_comp_in,dt_in,nt_in,fp_in,abc_in,lame_in,snap_in,write_bkg_in,absorbingBound_in,freeSurface_in,m1_in,m2_in,m3_in,b_in,e_in)

    integer             :: nx_in,nz_in,ntseis_in,rec_length_in,source_type_in,abc_in,data_comp_in
    integer*8			:: nt_in
    real                :: dx_in,dz_in,dtseis_in,zrec_in,xrec_in,xsource_in,zsource_in,dt_in,fp_in
    logical             :: adj_in,rec_list_in,lame_in,snap_in,write_bkg_in,absorbingBound_in,freeSurface_in
    !Background properties
    real, dimension(:,:),target  :: m1_in,m2_in,m3_in
    integer, dimension(:), target :: b_in, e_in

	adj = adj_in
	!Model dimensions
    nx = nx_in
    dx = dx_in
    nz = nz_in
    dz = dz_in
    !Data dimensions
    dtseis = dtseis_in
    ntseis = ntseis_in
    nrec = rec_length_in
	rec_list = rec_list_in
	!Receiver infor
    zrec = zrec_in
    xrec = xrec_in
    data_comp = data_comp_in
	!Source info
    xsource = xsource_in
    zsource = zsource_in
    source_type  = source_type_in
	!Propagation info
    dt = dt_in
    nt = nt_in
    fp = fp_in
    abc = abc_in
    lame = lame_in
    snap = snap_in
    write_bkg = write_bkg_in
	absorbingBound = absorbingBound_in
	freeSurface = freeSurface_in
    !Input background model
    m1 => m1_in
    m2 => m2_in
    m3 => m3_in

	!Nodes OMP info
    b => b_in
    e => e_in
    nodes = size(b)

    !Add border to receiver and source real coordinates
    xsource = xsource + (abc*dx)
	if(freeSurface) then
	  zsource = zsource + (4*dx)
	  zrec    = zrec + (4*dx)
	else
      zsource = zsource + (abc*dx)
      zrec = zrec + (abc*dx)
	endif
    
    !Allocate model parameters
    allocate(m1_x(nz,nx))
    allocate(m1_z(nz,nx))

    allocate(aux_m1_x(nz,nx))
    allocate(aux_m1_z(nz,nx))
    allocate(aux_m2(nz,nx))
    allocate(aux_m3(nz,nx))
    allocate(aux_m2m3(nz,nx))

    !Allocate source
    allocate(source(nt))
    allocate(source_wavelet(ntseis))
    source_wavelet=0
    source=0.  

	write(0,*)nt,size(source),ntseis,size(source_wavelet)

     
    !Allocate receiver list
    !Z Coordinate (real)
    izrec = (zrec/dx)
    !X Coordinate (real)
    if(rec_list) then
      allocate (rec_coords(nrec))
      call sep_read(rec_coords,"rec_file")
      rec_coords = rec_coords + (abc*dx)
      write(0,*)"READING RECEIVER POSITIONS FROM LIST ==> OK"
	else
      allocate (rec_coords(nrec))
	  do irec=1,nrec
	    rec_coords(irec) = ((irec-1)+abc)*dx + xrec
	  enddo
	  write(0,*)"RECEIVER POSITION AT ALL X"
	  write(0,*)"FIRST RECEIVER AT (",rec_coords(1),",",zrec,")"
	  write(0,*)"TOTAL RECEIVERS:",nrec
	endif

    !Data    
 	long1=5*nz*nx
    allocate(waveNew(long1))
    allocate(waveCur(long1))

	!Receiver wavefields
	waveNew = 0.
	waveCur = 0.
    pwaveNew => waveNew
    pwaveCur => waveCur

!     Write outputs to the terminal
    write(0,*)"============================================================="
    write(0,*)"SOURCE POSITION AFTER ADDING BOUNDARY"
    write(0,*)"============================================================="
    write(0,*)"(zsource,xsource)=",zsource,xsource
    write(0,*)"============================================================="
    write(0,*)"MEMORY ALLOCATION FOR TEMPORARY ARRAYS"
    write(0,*)"============================================================="
    write(0,*)"PROPERTY ARRAY"
    write(0,*)"m1",size(m1)*4/1024,"KB"
    write(0,*)"m2",size(m2)*4/1024,"KB"
    write(0,*)"m3",size(m3)*4/1024,"KB"
    write(0,*)"INTERPOLATED DENSITY"
    write(0,*)"rho_x",size(m1_x)*4/1024,"KB"
    write(0,*)"rho_z",size(m1_z)*4/1024,"KB"
    write(0,*)"TEMPORARY ARRAYS"
    write(0,*)"waveNew",size(waveNew)*4/1024,"KB"
    write(0,*)"waveCur",size(waveCur)*4/1024,"KB"    
    write(0,*)"AUXILIARY PROPERTY ARRAYS"
    write(0,*)"aux_m1_x",size(aux_m1_x)*4/1024,"KB"
    write(0,*)"aux_m1_z",size(aux_m1_z)*4/1024,"KB"
    write(0,*)"aux_m2",size(aux_m2)*4/1024,"KB"
    write(0,*)"aux_m3",size(aux_m3)*4/1024,"KB"
    write(0,*)"aux_m2m3",size(aux_m2m3)*4/1024,"KB"
    write(0,*)"============================================================="

    !Initialize values
    !Background properties
    m1_x  = 0.
    m1_z  = 0.

    !Initialize subroutines
    !Model
    call model_init(nz,nx,dx)
    call model_op(m1,m2,m3,m1_x,m1_z)
    write(0,*)"MODEL_OP => OK"
    !Absorbing boundary
    if (absorbingBound) call abc_init(nz,nx,dx,abc,freeSurface)
    write(0,*)"ABC_INIT => OK"
    !Snapshot
    if(snap) call snap_init(nz,dz,nx,dx,ntseis,dtseis)
    
    aux = dt/dx
    aux2 = sqrt(float(ntseis)/float(nt))
    aux3= 1.0/dx
    aux4= 1.0/dt
    dt2 = dt/2.0 !The do loop actually steps at half time steps,
                 !so this is needed to sample the seismogram correctly
	dtratio = ceiling(dtseis/dt)

    !Define sinc length
    !nf=11
    nf=5
    nm=(nf-1)/2    
    !Allocate sinc filter
    allocate(filter(nf,nf))
	!Sinc
	!call sinc_init(0.1,nf,filter)
	call gauss_init(0.8,nf,filter)
	!call spike_init(0.8,nf,filter)

    aux_m1_x = aux*m1_x
    aux_m1_z = aux*m1_z
    if (lame) then
      aux_m2 = aux*1.0E6*m2
      aux_m3 = aux*1.0E6*m3
      aux_m2m3 = aux*(1.0E6*m2+2.0*1.0E6*m3)
    else
      aux_m2 = aux*m1*(m2*m2-2.0*m3*m3)
      aux_m3 = aux*m1*m3*m3
      aux_m2m3 = aux*m1*m2*m2
    endif

    !Source
    call source_init(fp,dt2,dx,nz,nx,nt,ntseis,dtseis,abc,aux_m1_x,aux_m1_z)
    if(adj==.false.) call source_op2(source_wavelet)
    write(0,*)"SOURCE_OP => OK"

  end subroutine

! Function that is called by Main
integer function elastic_op(rec)
    real,dimension(:),target      :: rec

    call elastic_op2(rec)
	elastic_op = 0

  end function

! Subroutine used by function
subroutine elastic_op2(rec)
    real,dimension(:),target :: rec

    !General variables
    integer                         :: it, ix, iz,ierr

    pRec    => rec

	if(adj==.false.) then	
!###################################################
! FORWARD NON-LINEAR
!###################################################
		ierr = source_interp(F,T,source_wavelet,source)
		write(0,*)"SOURCE_INTERP => OK"	

		write(0,*)"FORWARD NON-LINEAR"
		write(0,*)"Processing:"
		write(0,*)"0%           25%            50%           75%            100%"
		write(0,'(a)',advance="no")"["
		
		do it=1,nt

		  ierr = source_op(F,T,it,source,pwaveNew)
		  ierr = propagate(F,T,pwaveCur,pwaveNew)
		  if(absorbingBound) ierr = absorbing_bound(pwaveNew)
		  if(snap) ierr = snapshot(it,pwaveNew)
		  ierr = receiver(F,T,it,pwaveNew,pRec)

		  pwaveTemp => pwaveCur
		  pwaveCur  => pwaveNew
		  pwaveNew  => pwaveTemp

		  !Write to terminal
		  if (mod(it,(nt/59))==0) then
			write(0,'(a)',advance="no")"#"
			close(0)
		  endif			  
		enddo
		write(0,'(a)',advance="yes")"]"
		if(write_bkg) call bkg_wave(.false.,nx,dx,nz,dx,pwaveCur,pwaveNew) 
 
	else
!###################################################
! ADJOINT LINEAR WITH RESPECT TO SOURCE
!###################################################
		write(0,*)"FORWARD NON-LINEAR"
		write(0,*)"Processing:"
		write(0,*)"0%           25%            50%           75%            100%"
		write(0,'(a)',advance="no")"["

		source = 0.

		do it=nt,1,-1
		  ierr = receiver(T,T,it,pwaveNew,pRec)
		  if(snap) ierr = snapshot(it,pwaveNew)
		  if(absorbingBound) ierr = absorbing_bound(pwaveNew)
		  ierr = propagate(T,T,pwaveCur,pwaveNew)
		  ierr = source_op(T,T,it,source,pwaveNew)

		  pwaveTemp => pwaveCur
		  pwaveCur  => pwaveNew
		  pwaveNew  => pwaveTemp

		  !Write to terminal
		  if (mod(it,(nt/59))==0) then
			write(0,'(a)',advance="no")"#"
			close(0)
		  endif
		enddo

		write(0,'(a)',advance="yes")"]"
		ierr = source_interp(T,T,source_wavelet,source)
 		write(0,*)"SOURCE_INTERP => OK"	

 	endif   
!###################################################
! BACKPROPAGATION NON-LINEAR
!###################################################
! 	write(0,*)"BACKWARD NON-LINEAR"
!     write(0,*)"Processing:"
!     write(0,*)"0%           25%            50%           75%            100%"
!     write(0,'(a)',advance="no")"["
! 
! 	do it=nt,1,-1
! 
! 	  ierr = depropagate(F,T,pwaveCur,pwaveNew)
! 	  if(snap) ierr = snapshot_bkg(it,pwaveNew)
! 
!  	  pwaveTemp => pwaveCur
!  	  pwaveCur  => pwaveNew
!  	  pwaveNew  => pwaveTemp
! 
!       !Write to terminal
!       if (mod(it,(nt/59))==0) then
!         write(0,'(a)',advance="no")"#"
!         close(0)
!       endif
!       	  	  
! 	enddo
!     write(0,'(a)',advance="yes")"]"
                 
end subroutine

!###################################################
! FUNCTIONS THAT ARE CALLED BY MAIN
!###################################################

integer function propagate(adj,add,model,dat)
   integer,intent(in) :: adj,add
   real               :: model(:),dat(:)
   
   call propagate2d(adj,add,model(1:nx*nz),model(1+nx*nz:2*nx*nz),model(1+2*nx*nz:3*nx*nz),model(1+3*nx*nz:4*nx*nz),model(1+4*nx*nz:5*nx*nz),dat(1:nx*nz),dat(1+nx*nz:2*nx*nz),dat(1+2*nx*nz:3*nx*nz),dat(1+3*nx*nz:4*nx*nz),dat(1+4*nx*nz:5*nx*nz))
   propagate=0

end function

integer function snapshot(it,dat)
	integer,intent(in)  :: it
	real				:: dat(:)

    if(mod(it,(2*dtratio)).eq.0) call snap_op5(dat(1:nx*nz),dat(1+nx*nz:2*nx*nz),dat(1+2*nx*nz:3*nx*nz),dat(1+3*nx*nz:4*nx*nz),dat(1+4*nx*nz:5*nx*nz))
	snapshot=0

end function

integer function snapshot_bkg(it,dat)
	integer,intent(in)  :: it
	real				:: dat(:)

    if(mod(it,(2*dtratio)).eq.0) call snap_op4(dat(1:nx*nz),dat(1+nx*nz:2*nx*nz),dat(1+2*nx*nz:3*nx*nz),dat(1+3*nx*nz:4*nx*nz),dat(1+4*nx*nz:5*nx*nz))
	snapshot_bkg=0

end function

integer function receiver(adj,add,it,model,dat)
   integer,intent(in) :: adj,add,it
   real			      :: model(:),dat(:)

   if(data_comp.le.1) then
     call receiver2d(adj,add,it,model(1:nx*nz),model(1+nx*nz:2*nx*nz),model(1+2*nx*nz:3*nx*nz),model(1+3*nx*nz:4*nx*nz),model(1+4*nx*nz:5*nx*nz),dat(1:nrec*ntseis),dat(1+nrec*ntseis:2*nrec*ntseis),dat(1+2*nrec*ntseis:3*nrec*ntseis),dat(1+3*nrec*ntseis:4*nrec*ntseis),dat(1+4*nrec*ntseis:5*nrec*ntseis))
   else
     call receiver2df(adj,add,it,model(1:nx*nz),model(1+nx*nz:2*nx*nz),model(1+2*nx*nz:3*nx*nz),model(1+3*nx*nz:4*nx*nz),model(1+4*nx*nz:5*nx*nz),dat(1:nrec*ntseis),dat(1+nrec*ntseis:2*nrec*ntseis),dat(1+2*nrec*ntseis:3*nrec*ntseis),dat(1+3*nrec*ntseis:4*nrec*ntseis),dat(1+4*nrec*ntseis:5*nrec*ntseis))
   endif
   receiver=0

end function

integer function absorbing_bound(model)
	real			   :: model(:)
	
	call K(model(1        :  nx*nz))
    call K(model(1+  nx*nz:2*nx*nz))
    call K(model(1+2*nx*nz:3*nx*nz))
    call K(model(1+3*nx*nz:4*nx*nz))
    call K(model(1+4*nx*nz:5*nx*nz))

	absorbing_bound=0
	
end function

integer function source_op(adj,add,it,model,dat)
	integer	:: adj,add, it
	real	:: model(:),dat(:)

    call source_op2d(adj,add,it,model,dat(1:nx*nz),dat(1+nx*nz:2*nx*nz),dat(1+2*nx*nz:3*nx*nz),dat(1+3*nx*nz:4*nx*nz))

	
! 	call source_op2d(adj,add,it,model(1:nt),dat(1:nx*nz),dat(1+nx*nz:2*nx*nz),dat(1+2*nx*nz:3*nx*nz),dat(1+3*nx*nz:4*nx*nz))
	
	source_op=0
	
end function

integer function source_interp(adj,add,model,data)
	integer	:: adj,add
	real	:: model(:),data(:)

	call source_interp2d(adj,add,model,data)

	source_interp=0
	
end function

!TESTING BACK PROPAGATION
integer function depropagate(adj,add,model,dat)
   integer,intent(in) :: adj,add
   real               :: model(:),dat(:)
   
   call depropagate2d(adj,add,model(1:nx*nz),model(1+nx*nz:2*nx*nz),model(1+2*nx*nz:3*nx*nz),model(1+3*nx*nz:4*nx*nz),model(1+4*nx*nz:5*nx*nz),dat(1:nx*nz),dat(1+nx*nz:2*nx*nz),dat(1+2*nx*nz:3*nx*nz),dat(1+3*nx*nz:4*nx*nz),dat(1+4*nx*nz:5*nx*nz))
   depropagate=0

end function

!###################################################
! SUBROUTINES
!###################################################

subroutine propagate2d(adj,add,Vx_cur,Vz_cur,sigmaxx_cur,sigmazz_cur,sigmaxz_cur, &
							   Vx_new,Vz_new,sigmaxx_new,sigmazz_new,sigmaxz_new)
  integer 				:: adj,add
  real,dimension(nz,nx) :: Vx_cur, Vz_cur, Vx_new, Vz_new
  real,dimension(nz,nx) :: sigmaxz_cur, sigmazz_cur, sigmaxx_cur
  real,dimension(nz,nx) :: sigmaxz_new, sigmazz_new, sigmaxx_new
  integer :: ix   

    b(1)=5
    e(nodes)=nx-5
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(adj,add,nx,nz,b(ix),e(ix),sigmaxz_cur,Vx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(adj,add,nx,nz,b(ix),e(ix),Vz_cur,sigmaxx_new,aux_m2)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bx_c(adj,add,nx,nz,b(ix),e(ix),Vz_cur,sigmazz_new,aux_m2m3)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(adj,add,nx,nz,b(ix),e(ix),sigmazz_cur,Vz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bz_c(adj,add,nx,nz,b(ix),e(ix),Vx_cur,sigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO
   
    if (adj == T) b(1)=9
    if (adj == T) e(nodes)=nx-10   
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(adj,add,nx,nz,b(ix),e(ix),sigmaxx_cur,Vx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Ax_c(adj,add,nx,nz,b(ix),e(ix),Vz_cur,sigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO

    if (adj == T) b(1)=10
    if (adj == T) e(nodes)=nx-9   
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(adj,add,nx,nz,b(ix),e(ix),sigmaxz_cur,Vz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(adj,add,nx,nz,b(ix),e(ix),Vx_cur,sigmaxx_new,aux_m2m3)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Az_c(adj,add,nx,nz,b(ix),e(ix),Vx_cur,sigmazz_new,aux_m2)
    enddo
!$OMP END PARALLEL DO

	if(adj == T) then
!$OMP SECTIONS
!$OMP SECTION	
      call Azsides_c(T,nx,nz,Vx_cur,sigmaxx_new,aux_m2m3)
      call Azsides_c(T,nx,nz,Vx_cur,sigmazz_new,aux_m2)  
      call Bzsides_c(T,nx,nz,Vx_cur,sigmaxz_new,aux_m3)
!$OMP SECTION	
      call Bxsides_c(T,nx,nz,Vz_cur,sigmaxx_new,aux_m2)
      call Bxsides_c(T,nx,nz,Vz_cur,sigmazz_new,aux_m2m3)
      call Axsides_c(T,nx,nz,Vz_cur,sigmaxz_new,aux_m3)
!$OMP SECTION	
      call Axsides_c(T,nx,nz,sigmaxx_cur,Vx_new,aux_m1_x)
!$OMP SECTION	
      call Bzsides_c(T,nx,nz,sigmazz_cur,Vz_new,aux_m1_z)
!$OMP SECTION
      call Bxsides_c(T,nx,nz,sigmaxz_cur,Vx_new,aux_m1_x)
      call Azsides_c(T,nx,nz,sigmaxz_cur,Vz_new,aux_m1_z)
!$OMP END SECTIONS
	endif

end subroutine

subroutine receiver2d(adj,add,it,Vx_new,Vz_new,sigmaxx_new,sigmazz_new,sigmaxz_new, Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  integer 				:: adj,add,it
  real,dimension(nz,nx) :: Vx_new, Vz_new
  real,dimension(nz,nx) :: sigmaxx_new, sigmazz_new, sigmaxz_new
  real,dimension(nrec,ntseis) :: Vx_rec, Vz_rec, sigmaxx_rec, sigmazz_rec, sigmaxz_rec

  do irec=1,nrec
    xrec = rec_coords(irec)
    ixrec = int(xrec/dx)
	itseis=int(it*(dt2/dtseis))
    
!$OMP SECTIONS
!$OMP SECTION
    call sinc_serial2_c(adj,add,nf,nz,nx,xrec,zrec,dx,it,dt2,dtseis,ntseis,irec,nrec,aux2,Vx_new,filter,Vx_rec)
!$OMP SECTION
    call sinc_serial2_c(adj,add,nf,nz,nx,xrec,zrec,dx,it,dt2,dtseis,ntseis,irec,nrec,aux2,Vz_new,filter,Vz_rec)
!$OMP SECTION
    call sinc_serial2_c(adj,add,nf,nz,nx,xrec,zrec,dx,it,dt2,dtseis,ntseis,irec,nrec,aux2,sigmaxx_new,filter,sigmaxx_rec)
!$OMP SECTION
    call sinc_serial2_c(adj,add,nf,nz,nx,xrec,zrec,dx,it,dt2,dtseis,ntseis,irec,nrec,aux2,sigmazz_new,filter,sigmazz_rec)
!$OMP END SECTIONS
  enddo

end subroutine

subroutine receiver2df(adj,add,it,Vx_new,Vz_new,sigmaxx_new,sigmazz_new,sigmaxz_new, Vx_rec,Vz_rec,sigmaxx_rec,sigmazz_rec,sigmaxz_rec)
  integer 				:: adj,add,it
  real,dimension(nz,nx) :: Vx_new, Vz_new
  real,dimension(nz,nx) :: sigmaxx_new, sigmazz_new, sigmaxz_new
  real,dimension(nrec,ntseis) :: Vx_rec, Vz_rec, sigmaxx_rec, sigmazz_rec, sigmaxz_rec

  do irec=1,nrec
    xrec = rec_coords(irec)
    ixrec = int(xrec/dx)
	itseis=int(it*(dt2/dtseis))
    
!$OMP SECTIONS
!$OMP SECTION
    call sinc_serial2_fx_c(adj,add,nf,nz,nx,xrec,zrec,dx,it,dt2,dtseis,ntseis,irec,nrec,aux2,Vx_new,filter,Vx_rec)
!$OMP SECTION
    call sinc_serial2_fz_c(adj,add,nf,nz,nx,xrec,zrec,dx,it,dt2,dtseis,ntseis,irec,nrec,aux2,Vz_new,filter,Vz_rec)
!$OMP END SECTIONS
  enddo

end subroutine

subroutine source_interp2d(adj,add,source_wavelet,source)
	integer			:: adj,add
	integer			:: wstat,sseek,srite
	real,dimension(ntseis)  :: source_wavelet
	real,dimension(nt)		:: source

	  call source_interp1d(adj,add,source_wavelet,source)
	  if (adj) then
	  	call to_history("n1",ntseis,"source_wavelet")
		call to_history("d1",dtseis,"source_wavelet")
	  	call to_history("o1",0.,"source_wavelet")
		wstat = sseek("source_wavelet",0,0)
	  	wstat = srite("source_wavelet",source_wavelet,4*ntseis)
    	if (wstat.eq.4*ntseis) then
    		write(0,*)"WRITE INVERTED SOURCE => OK"
    	else
    		write(0,*)"WRITE INVERTED SOURCE => FAIL"
    		stop
    	endif
	  else
! 	  	call to_history("n1",nt,"test_source_interp.H")
! 		call to_history("d1",dt2,"test_source_interp.H")
! 	  	call to_history("o1",0.,"test_source_interp.H")
! 	  	call sep_write(source,"test_source_interp.H")
	  endif

end subroutine

subroutine source_op2d(adj,add,it,source,Vx_new,Vz_new,sigmaxx_new,sigmazz_new)
  integer				:: adj,add,it
  real,dimension(nt) 	:: source
  real,dimension(nz,nx) :: Vx_new, Vz_new
  real,dimension(nz,nx) :: sigmaxx_new, sigmazz_new

	if(source_type.eq.1) then
      call explosive_source(adj,add,source(it),xsource,zsource,sigmaxx_new,sigmazz_new)
	endif
	if(source_type.eq.2) call fz_source(adj,add,source(it),xsource,zsource,Vz_new)
	if(source_type.eq.3) call fx_source(adj,add,source(it),xsource,zsource,Vx_new)
	if(source_type.eq.4) call fxfz_source(adj,add,source(it),xsource,zsource,Vx_new,Vz_new)
 
end subroutine

subroutine depropagate2d(adj,add,Vx_cur,Vz_cur,sigmaxx_cur,sigmazz_cur,sigmaxz_cur, &
							 Vx_new,Vz_new,sigmaxx_new,sigmazz_new,sigmaxz_new)
  integer 				:: adj,add
  real,dimension(nz,nx) :: Vx_cur, Vz_cur, Vx_new, Vz_new
  real,dimension(nz,nx) :: sigmaxz_cur, sigmazz_cur, sigmaxx_cur
  real,dimension(nz,nx) :: sigmaxz_new, sigmazz_new, sigmaxx_new
  integer :: ix 

    b(1)=5
    e(nodes)=nx-5
    !Vx
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Axb_c(adj,add,nx,nz,b(ix),e(ix),sigmaxx_cur,Vx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bxb_c(adj,add,nx,nz,b(ix),e(ix),sigmaxz_cur,Vx_new,aux_m1_x)
    enddo
!$OMP END PARALLEL DO
    !Vz
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Azb_c(adj,add,nx,nz,b(ix),e(ix),sigmaxz_cur,Vz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
      call Bzb_c(adj,add,nx,nz,b(ix),e(ix),sigmazz_cur,Vz_new,aux_m1_z)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Azb_c(adj,add,nx,nz,b(ix),e(ix),Vx_cur,sigmaxx_new,aux_m2m3)
    !sigmazz
      call Azb_c(adj,add,nx,nz,b(ix),e(ix),Vx_cur,sigmazz_new,aux_m2)
    !sigmaxz
      call Bzb_c(adj,add,nx,nz,b(ix),e(ix),Vx_cur,sigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO
!$OMP PARALLEL DO DEFAULT(SHARED) PRIVATE(IX)
    do ix=1,nodes
    !sigmaxx
      call Bxb_c(adj,add,nx,nz,b(ix),e(ix),Vz_cur,sigmaxx_new,aux_m2)
    !sigmazz
      call Bxb_c(adj,add,nx,nz,b(ix),e(ix),Vz_cur,sigmazz_new,aux_m2m3)
    !sigmaxz
      call Axb_c(adj,add,nx,nz,b(ix),e(ix),Vz_cur,sigmaxz_new,aux_m3)
    enddo
!$OMP END PARALLEL DO
  
end subroutine 
  
end module
