! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! seismogram module
module image_mod_rtm

  use sep
  !use omp_lib

  implicit none
  integer, private :: nx,nz,nt,ix,iz
  real, private    :: dx,dz,dt
  real, private    :: c1=35.0/294912.0,c2=-405.0/229376.0,c3=567.0/40960.0,c4=-735.0/8192.0,c5=19845.0/16384.0
  real*8, dimension(:,:,:), pointer, private :: Vx0,Vz0
  real*8, dimension(:,:), pointer, private   :: m1,m2,m3
  real*8, allocatable, dimension(:,:), private :: S1,S2,S3,S4,S5,S6
  logical, private :: lame

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine image_init(nx_in,dx_in,nz_in,dz_in,nt_in,dt_in,Vx0_in,Vz0_in,lame_in)
    integer,intent(in)              :: nx_in,nz_in,nt_in
    real,intent(in)                 :: dx_in,dz_in,dt_in
    logical                         :: lame_in
    real*8,dimension(:,:,:),target  :: Vx0_in,Vz0_in

    nx = nx_in
    dx = dx_in
    nz = nz_in
    dz = dz_in
    nt = nt_in
    dt = dt_in
    lame = lame_in

    Vx0 => Vx0_in
    Vz0 => Vz0_in

    allocate(S1(nz,nx))
    allocate(S2(nz,nx))
    allocate(S3(nz,nx))
    allocate(S4(nz,nx))
    allocate(S5(nz,nx))
    allocate(S6(nz,nx))

    S1 = 0.
    S2 = 0.
    S3 = 0.
    S4 = 0.
    S5 = 0.
    S6 = 0.

    !call T(.false.,.false.,Vx0,S1)
    !call T(.false.,.false.,Vz0,S2)
    !call vxdx(.false.,.false.,Vx0,S4)
    !call vzdz(.false.,.false.,Vz0,S5)
    !S3=S4+S5
    !call vxdz(.false.,.false.,Vx0,S6)
    !call vzdx(.false.,.true.,Vz0,S6)

  endsubroutine

  subroutine image_init2(m1_in,m2_in,m3_in)
    real*8,dimension(:,:),target  :: m1_in,m2_in,m3_in

    m1 => m1_in
    m2 => m2_in
    m3 => m3_in

  endsubroutine

!==================================
! FUNCTION CALLS
!==================================

  function image_op(adj,add,dm1,dm1_x,dm1_z,dm2,dm3,Vx,Vz,sigmaxx,sigmazz,sigmaxz,it) result(stat)
    logical,intent(in)          :: adj,add
    real*8,dimension(:)         :: dm1,dm1_x,dm1_z,dm2,dm3,Vx,Vz,sigmaxx,sigmazz,sigmaxz
    integer                     :: stat,it

    call image_op2(adj,add,dm1,dm1_x,dm1_z,dm2,dm3,Vx,Vz,sigmaxx,sigmazz,sigmaxz,it)
    stat=0
  endfunction

  function funT(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real*8,dimension(:)          :: model,data
    integer                      :: stat
    call T(adj,add,model,data)
    stat=0
  endfunction

  function funvxdx(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real*8,dimension(:)          :: model,data
    integer                      :: stat
    call vxdx(adj,add,model,data)
    stat=0
  endfunction

  function funvzdz(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real*8,dimension(:)          :: model,data
    integer                      :: stat
    call vzdz(adj,add,model,data)
    stat=0
  endfunction

  function funvxdz(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real*8,dimension(:)          :: model,data
    integer                      :: stat
    call vxdz(adj,add,model,data)
    stat=0
  endfunction

  function funvzdx(adj,add,model,data) result(stat)
    logical,intent(in)           :: adj,add
    real*8,dimension(:)          :: model,data
    integer                      :: stat
    call vzdx(adj,add,model,data)
    stat=0
  endfunction

!==================================
! SUBROUTINES
!==================================
   subroutine image_op2(adj,add,dm1,dm1_x,dm1_z,dm2,dm3,Vx,Vz,sigmaxx,sigmazz,sigmaxz,it)
    logical,intent(in)                       :: adj,add
    integer                                  :: it
    real*8,dimension(nz,nx),intent(inout)    :: dm1,dm1_x,dm1_z,dm2,dm3
    real*8,dimension(nz,nx),intent(inout)    :: Vx,Vz,sigmaxx,sigmazz,sigmaxz

    call T(.false.,.false.,Vx0(:,:,it:it+1),S1)
    call T(.false.,.false.,Vz0(:,:,it:it+1),S2)
    call vxdx(.false.,.false.,Vx0(:,:,it),S4)
    call vzdz(.false.,.false.,Vz0(:,:,it),S5)
    S3=S4+S5
    call vxdz(.false.,.false.,Vx0(:,:,it),S6)
    call vzdx(.false.,.true.,Vz0(:,:,it),S6)

! IMAGING CONDITION FOR LAME PARAMETERS
    if (lame) then
      if(adj)then
        if(.not.add) then
          dm1_x = 0.
          dm1_z = 0.
          dm2   = 0.
          dm3   = 0.
        endif
        dm1_x(:,:) = dm1_x(:,:) + (Vx(:,:)*S1(:,:))
        dm1_z(:,:) = dm1_z(:,:) + (Vz(:,:)*S2(:,:))
        dm2(:,:) = dm2(:,:) - (sigmaxx(:,:)*S3(:,:)) &
                            - (sigmazz(:,:)*S3(:,:))
        dm3(:,:) = dm3(:,:) - (sigmaxx(:,:)*2.0*S4(:,:)) &
                            - (sigmazz(:,:)*2.0*S5(:,:)) &
                            - (sigmaxz(:,:)*S6(:,:))
      else
        if(.not.add) then
          Vx = 0.
          Vz = 0.
          sigmaxx = 0.
          sigmazz = 0.
          sigmaxz = 0.
        endif
        Vx(:,:) = Vx(:,:) + (dm1_x(:,:)*S1(:,:))
        Vz(:,:) = Vz(:,:) + (dm1_z(:,:)*S2(:,:))
        sigmaxx(:,:) = sigmaxx(:,:) - (dm2(:,:)*S3(:,:)) &
                                    - (dm3(:,:)*2.0*S4(:,:))
        sigmazz(:,:) = sigmazz(:,:) - (dm2(:,:)*S3(:,:)) &
                                    - (dm3(:,:)*2.0*S5(:,:))
        sigmaxz(:,:) = sigmaxz(:,:) - (dm3(:,:)*S6(:,:))
      endif
! IMAGING CONDITION FOR VELOCITY PARAMETERS
    else
      if(adj)then
        if(.not.add) then
          dm1 = 0.
          dm1_x = 0.
          dm1_z = 0.
          dm2 = 0.
          dm3 = 0.
        endif
        dm1_x(:,:) = dm1_x(:,:) + (Vx(:,:)*S1(:,:))
        dm1_z(:,:) = dm1_z(:,:) + (Vz(:,:)*S2(:,:))
        dm1(:,:) = dm1(:,:) - (sigmaxx(:,:)*(m2(:,:)*m2(:,:)*S4(:,:) &
                                           +(m2(:,:)*m2(:,:)-2.0*m3(:,:)*m3(:,:))*S5(:,:))) &
                            - (sigmazz(:,:)*(m2(:,:)*m2(:,:)*S5(:,:) &
                                           +(m2(:,:)*m2(:,:)-2.0*m3(:,:)*m3(:,:))*S4(:,:))) &
                            - (sigmaxz(:,:)*(m3(:,:)*m3(:,:)*S6(:,:)))
        dm2(:,:) = dm2(:,:) - (sigmaxx(:,:)*(2.0*m1(:,:)*m2(:,:)*S3(:,:))) &
                            - (sigmazz(:,:)*(2.0*m1(:,:)*m2(:,:)*S3(:,:)))
        dm3(:,:) = dm3(:,:) + (sigmaxx(:,:)*(4.0*m1(:,:)*m3(:,:)*S5(:,:))) &
                            + (sigmazz(:,:)*(4.0*m1(:,:)*m3(:,:)*S4(:,:))) &
                            - (sigmaxz(:,:)*(2.0*m1(:,:)*m3(:,:)*S6(:,:)))
      else
        if(.not.add) then
          Vx = 0.
          Vz = 0.
          sigmaxx = 0.
          sigmazz = 0.
          sigmaxz = 0.
        endif
        Vx(:,:) = Vx(:,:) + (dm1_x(:,:)*S1(:,:))
        Vz(:,:) = Vz(:,:) + (dm1_z(:,:)*S2(:,:))
        sigmaxx(:,:) = sigmaxx(:,:) - (dm1(:,:)*(m2(:,:)*m2(:,:)*S4(:,:) &
                                               +(m2(:,:)*m2(:,:)-2.0*m3(:,:)*m3(:,:))*S5(:,:))) &
                                     - (dm2(:,:)*(2.0*m1(:,:)*m2(:,:)*S3(:,:))) &
                                     + (dm3(:,:)*(4.0*m1(:,:)*m3(:,:)*S5(:,:)))
        sigmazz(:,:) = sigmazz(:,:) - (dm1(:,:)*(m2(:,:)*m2(:,:)*S5(:,:) &
                                               +(m2(:,:)*m2(:,:)-2.0*m3(:,:)*m3(:,:))*S4(:,:))) &
                                    - (dm2(:,:)*(2.0*m1(:,:)*m2(:,:)*S3(:,:))) &
                                    + (dm3(:,:)*(4.0*m1(:,:)*m3(:,:)*S4(:,:)))
        sigmaxz(:,:) = sigmaxz(:,:) - (dm1(:,:)*(m3(:,:)*m3(:,:)*S6(:,:))) &
                                    - (dm3(:,:)*(2.0*m1(:,:)*m3(:,:)*S6(:,:)))
      endif
    endif
    
  endsubroutine
!==================================
! DERIVATIVES
!==================================

  subroutine T(adj,add,model,data)
    real                                     :: aux
    logical,intent(in)                       :: adj,add
    real*8,dimension(nz,nx),intent(inout)    :: data
    real*8,dimension(nz,nx,2),intent(inout)  :: model

    aux=1.0/dt

    if(adj) then
      !DONT BOTHER PASSING THE DOT PRODUCT TEST
      !YOU WILL NOT NEED IT!
      if(.not.add) model = 0.
      do ix=1,nx
        do iz=1,nz
          model(iz,ix,2) = model(iz,ix,2) + aux*data(iz,ix)
          model(iz,ix,1) = model(iz,ix,1) - aux*data(iz,ix)
        enddo
      enddo
    else
      if(.not.add) data = 0.
      do ix=1,nx
        do iz=1,nz
          data(iz,ix) = data(iz,ix) + aux*(model(iz,ix,2)-model(iz,ix,1))
        enddo
      enddo
    endif
  endsubroutine

  subroutine vxdx(adj,add,model,data)
    real                                  :: aux
    logical,intent(in)                    :: adj,add
    real*8,dimension(nz,nx),intent(inout) :: model,data

    aux=1.0/dx

    if(adj) then
      if(.not.add) model = 0.
      do ix=5,nx-5
        do iz=5,nz-5
          model(iz,ix+5) = model(iz,ix+5) + c1*data(iz,ix)*aux
          model(iz,ix+4) = model(iz,ix+4) + c2*data(iz,ix)*aux
          model(iz,ix+3) = model(iz,ix+3) + c3*data(iz,ix)*aux
          model(iz,ix+2) = model(iz,ix+2) + c4*data(iz,ix)*aux
          model(iz,ix+1) = model(iz,ix+1) + c5*data(iz,ix)*aux
          model(iz,ix  ) = model(iz,ix  ) - c5*data(iz,ix)*aux
          model(iz,ix-1) = model(iz,ix-1) - c4*data(iz,ix)*aux
          model(iz,ix-2) = model(iz,ix-2) - c3*data(iz,ix)*aux
          model(iz,ix-3) = model(iz,ix-3) - c2*data(iz,ix)*aux
          model(iz,ix-4) = model(iz,ix-4) - c1*data(iz,ix)*aux
        enddo
      enddo
    else
      if(.not.add) data = 0.
      do ix=5,nx-5
        do iz=5,nz-5
          data(iz,ix) = data(iz,ix) + (c1*(model(iz,ix+5) - model(iz,ix-4))+&
                                       c2*(model(iz,ix+4) - model(iz,ix-3))+&
                                       c3*(model(iz,ix+3) - model(iz,ix-2))+&
                                       c4*(model(iz,ix+2) - model(iz,ix-1))+&
                                       c5*(model(iz,ix+1) - model(iz,ix  )))*aux
        enddo
      enddo
    endif

  endsubroutine

  subroutine vzdz(adj,add,model,data)
    real                                   :: aux
    logical,intent(in)                     :: adj,add
    real*8,dimension(nz,nx),intent(inout)  :: model,data

    aux=1.0/dx

    if(adj) then
      if(.not.add) model = 0.
      do ix=5,nx-5
        do iz=5,nz-5
          model(iz+5,ix) = model(iz+5,ix) + c1*data(iz,ix)*aux
          model(iz+4,ix) = model(iz+4,ix) + c2*data(iz,ix)*aux
          model(iz+3,ix) = model(iz+3,ix) + c3*data(iz,ix)*aux
          model(iz+2,ix) = model(iz+2,ix) + c4*data(iz,ix)*aux
          model(iz+1,ix) = model(iz+1,ix) + c5*data(iz,ix)*aux
          model(iz  ,ix) = model(iz  ,ix) - c5*data(iz,ix)*aux
          model(iz-1,ix) = model(iz-1,ix) - c4*data(iz,ix)*aux
          model(iz-2,ix) = model(iz-2,ix) - c3*data(iz,ix)*aux
          model(iz-3,ix) = model(iz-3,ix) - c2*data(iz,ix)*aux
          model(iz-4,ix) = model(iz-4,ix) - c1*data(iz,ix)*aux
        enddo
      enddo
    else
      if(.not.add) data = 0.
      do ix=5,nx-5
        do iz=5,nz-5
          data(iz,ix) = data(iz,ix) + (c1*(model(iz+5,ix) - model(iz-4,ix))+&
                                       c2*(model(iz+4,ix) - model(iz-3,ix))+&
                                       c3*(model(iz+3,ix) - model(iz-2,ix))+&
                                       c4*(model(iz+2,ix) - model(iz-1,ix))+&
                                       c5*(model(iz+1,ix) - model(iz  ,ix)))*aux
        enddo
      enddo
    endif
  endsubroutine

  subroutine vxdz(adj,add,model,data)
    real                                  :: aux
    logical,intent(in)                    :: adj,add
    real*8,dimension(nz,nx),intent(inout) :: model,data

    aux=1.0/dx

    if(adj) then
      if(.not.add) model = 0.
      do ix=6,nx-5
        do iz=6,nz-5
          model(iz+4,ix) = model(iz+4,ix) + c1*data(iz,ix)*aux
          model(iz+3,ix) = model(iz+3,ix) + c2*data(iz,ix)*aux
          model(iz+2,ix) = model(iz+2,ix) + c3*data(iz,ix)*aux
          model(iz+1,ix) = model(iz+1,ix) + c4*data(iz,ix)*aux
          model(iz  ,ix) = model(iz  ,ix) + c5*data(iz,ix)*aux
          model(iz-1,ix) = model(iz-1,ix) - c5*data(iz,ix)*aux
          model(iz-2,ix) = model(iz-2,ix) - c4*data(iz,ix)*aux
          model(iz-3,ix) = model(iz-3,ix) - c3*data(iz,ix)*aux
          model(iz-4,ix) = model(iz-4,ix) - c2*data(iz,ix)*aux
          model(iz-5,ix) = model(iz-5,ix) - c1*data(iz,ix)*aux
        enddo
      enddo
    else
      if(.not.add) data = 0.
      do ix=6,nx-5
        do iz=6,nz-5
          data(iz,ix) = data(iz,ix) + (c1*(model(iz+4,ix) - model(iz-5,ix))+&
                                       c2*(model(iz+3,ix) - model(iz-4,ix))+&
                                       c3*(model(iz+2,ix) - model(iz-3,ix))+&
                                       c4*(model(iz+1,ix) - model(iz-2,ix))+&
                                       c5*(model(iz  ,ix) - model(iz-1,ix)))*aux
        enddo
      enddo
    endif
  endsubroutine

  subroutine vzdx(adj,add,model,data)
    real                                  :: aux
    logical,intent(in)                    :: adj,add
    real*8,dimension(nz,nx),intent(inout) :: model,data

    aux=1.0/dx

    if(adj)then
      if(.not.add) model = 0.
      do ix=6,nx-5
        do iz=6,nz-5
          model(iz,ix+4) = model(iz,ix+4) + c1*data(iz,ix)*aux
          model(iz,ix+3) = model(iz,ix+3) + c2*data(iz,ix)*aux
          model(iz,ix+2) = model(iz,ix+2) + c3*data(iz,ix)*aux
          model(iz,ix+1) = model(iz,ix+1) + c4*data(iz,ix)*aux
          model(iz,ix  ) = model(iz,ix  ) + c5*data(iz,ix)*aux
          model(iz,ix-1) = model(iz,ix-1) - c5*data(iz,ix)*aux
          model(iz,ix-2) = model(iz,ix-2) - c4*data(iz,ix)*aux
          model(iz,ix-3) = model(iz,ix-3) - c3*data(iz,ix)*aux
          model(iz,ix-4) = model(iz,ix-4) - c2*data(iz,ix)*aux
          model(iz,ix-5) = model(iz,ix-5) - c1*data(iz,ix)*aux
      enddo
    enddo
    else
      if(.not.add) data = 0.
        do ix=6,nx-5
          do iz=6,nz-5
          data(iz,ix) = data(iz,ix) + (c1*(model(iz,ix+4) - model(iz,ix-5))+&
                                       c2*(model(iz,ix+3) - model(iz,ix-4))+&
                                       c3*(model(iz,ix+2) - model(iz,ix-3))+&
                                       c4*(model(iz,ix+1) - model(iz,ix-2))+&
                                       c5*(model(iz,ix  ) - model(iz,ix-1)))*aux
        enddo
      enddo
    endif
  endsubroutine

endmodule

