! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! seismogram module
module seis_mod_rtm

  use sep

  implicit none
  integer, private :: nx,nt,nz,ix,iz,it,abc,nrec,data_comp
  real, private    :: dx,dt,dz,scale
  integer*8,private :: long1
  real, dimension(:), pointer, private  :: rec_data,waveNew,waveCur,model
  real, dimension(:),allocatable,private   :: temparray

  contains

  subroutine seismo(nx_in,dx_in,nt_in,dt_in,rec_data_in,data_comp_in)
    integer,intent(in)           :: nx_in,nt_in,data_comp_in
    real,intent(in)              :: dx_in,dt_in
    real, dimension(:), target   :: rec_data_in
    
    integer :: wstat,srite,sseek,auxinout
    
    nx = nx_in
    dx = dx_in
    nt = nt_in
    dt = dt_in
    rec_data => rec_data_in
    data_comp = data_comp_in

	scale = 1.0!2.2222224E-04/(4040.0)  !delta t / ( density * delta x)

	wstat = auxinout("rec_data")
	
    call to_history("n1", nx, "rec_data")
    call to_history("o1", 0., "rec_data")
    call to_history("d1", dx, "rec_data")
    call to_history("n2", nt, "rec_data")
    call to_history("o2", 0., "rec_data")
    call to_history("d2", dt, "rec_data")
    call to_history("n3", 3, "rec_data")
    call to_history("o3", 0., "rec_data")
    call to_history("d3", 1, "rec_data")

	long1=nt*nx     
    if(data_comp==0) then
    	write(0,*)"WRITING ALL DATA COMPONENTS"
    	allocate(temparray(3*long1))
		temparray(1:long1)=0.5*(rec_data(2*long1+1:3*long1)+rec_data(3*long1+1:4*long1))    	
		temparray(long1+1:2*long1)=rec_data(long1+1:2*long1)
		temparray(2*long1+1:3*long1)=rec_data(1:long1)
		wstat = sseek("rec_data",0,0)
        wstat = srite("rec_data",temparray,4*3*long1)
        deallocate(temparray)
    endif
	if(data_comp==1) then
    	write(0,*)"WRITING HYDROPHONE RECIPROCAL DATA COMPONENT"
		allocate(temparray(long1))
		temparray=0.5*(rec_data(2*long1+1:3*long1)+rec_data(3*long1+1:4*long1))
		wstat = sseek("rec_data",0,0)
        wstat = srite("rec_data",temparray,4*long1)
        deallocate(temparray)
    endif
	if(data_comp==2) then
    	write(0,*)"WRITING Vz RECIPROCAL DATA COMPONENT"
		allocate(temparray(long1))
		temparray=0.5*scale*(rec_data(1:long1)+rec_data(long1+1:2*long1))
		wstat = sseek("rec_data",4*long1,0)
        wstat = srite("rec_data",temparray,4*long1)
        deallocate(temparray)
    endif
	if(data_comp==3) then
    	write(0,*)"WRITING Vx RECIPROCAL DATA COMPONENT"
		allocate(temparray(long1))
		temparray=0.5*scale*(rec_data(1:long1)+rec_data(long1+1:2*long1))
		wstat = sseek("rec_data",4*2*long1,0)
        wstat = srite("rec_data",temparray,4*long1)
    	deallocate(temparray)
    endif

  endsubroutine

  subroutine seismo_adj(nx_in,dx_in,nt_in,dt_in,rec_data_in)
    integer,intent(in)           :: nx_in,nt_in
    real,intent(in)              :: dx_in,dt_in
    real, dimension(:), target   :: rec_data_in
    
    integer :: wstat,sreed,sseek,auxinout
    
    nx = nx_in
    dx = dx_in
    nt = nt_in
    dt = dt_in
    rec_data => rec_data_in

	wstat = auxinout("rec_data")

	long1=nt*nx     
    write(0,*)"READING ALL DATA COMPONENTS"
    allocate(temparray(3*long1))
    !wstat = sseek("rec_data",0,0)
    wstat = sreed("rec_data",temparray,4*3*long1)
    if (wstat.eq.4*3*long1) then
    	write(0,*)"READ RECEIVER DATA => OK"
    else
    	write(0,*)"READ RECEIVER DATA => FAIL"
    	stop
    endif
    !call sep_read(temparray,"rec_data")
write(0,*) "SEISMO ADJ",sum(temparray)
    rec_data(1:long1)=temparray(2*long1+1:3*long1)
    rec_data(long1+1:2*long1)=temparray(long1+1:2*long1)
    rec_data(2*long1+1:3*long1)=0.5*temparray(1:long1)
    rec_data(3*long1+1:4*long1)=0.5*temparray(1:long1)
    deallocate(temparray)

  endsubroutine


  subroutine bkg_wave(adj,nx_in,dx_in,nz_in,dz_in,waveCur_in,waveNew_in)
    integer,intent(in)           :: nx_in,nz_in
    real,intent(in)              :: dx_in,dz_in
    real,dimension(:),target	 :: waveCur_in,waveNew_in
    logical,intent(in)			 :: adj
 
    nx = nx_in
    dx = dx_in
    nz = nz_in
    dz = dz_in
    
    waveCur => waveCur_in
    waveNew => waveNew_in

	long1 = nz*nx*5
	allocate(temparray(nz*nx*5*2))

	if(adj) then
		call sep_read(temparray,"bkg_wavefield")
	
		waveCur = temparray(1:5*nx*nz)
		waveNew = temparray(5*nx*nz+1:10*nx*nz)
	else
		temparray(1:5*nx*nz) = waveCur
		temparray(5*nx*nz+1:10*nx*nz) = waveNew

		call to_history("n1", nz, "bkg_wavefield")
		call to_history("o1", 0., "bkg_wavefield")
		call to_history("d1", dz, "bkg_wavefield")
		call to_history("n2", nx, "bkg_wavefield")
		call to_history("o2", 0., "bkg_wavefield")
		call to_history("d2", dx, "bkg_wavefield")
		call to_history("n3", 5,  "bkg_wavefield")
		call to_history("o3", 0., "bkg_wavefield")
		call to_history("d3", 1,  "bkg_wavefield")
		call to_history("n4", 2,  "bkg_wavefield")
		call to_history("o4", 0., "bkg_wavefield")    
		call to_history("d4", 1,  "bkg_wavefield")
		call sep_write(temparray, "bkg_wavefield")
	endif

    deallocate(temparray)
    
  endsubroutine

  subroutine img(nx_in,dx_in,nz_in,dz_in,model_in)
    integer,intent(in)          :: nx_in,nz_in
    real,intent(in)             :: dx_in,dz_in
    real, dimension(:), target  :: model_in

    nx = nx_in
    dx = dx_in
    nz = nz_in
    dz = dz_in
    model => model_in

    call to_history("n1", nz, "model")
    call to_history("o1", 0., "model")
    call to_history("d1", dz, "model")
    call to_history("n2", nx, "model")
    call to_history("o2", 0., "model")
    call to_history("d2", dx, "model")
    call to_history("n3",  3, "model")
    call to_history("o3", 0., "model")
    call to_history("d3",  1, "model")    
         
    call sep_write(model, "model")
    
  endsubroutine

endmodule
