! Linear interpolation operator module
module sinc_mod

  use sep

  implicit none
  real,pointer,dimension(:,:),private	:: filter
  integer, private						:: nf,nm,i1,i2
  real, private							:: aux, temp, fc, b

  contains

  subroutine sinc_init(fc_in,nf_in,filter_in)
    real, parameter      		:: pi=3.14159265359
    real						:: fc_in
    integer						:: nf_in
    real,dimension(:,:),target	:: filter_in

    fc = fc_in
    nf = nf_in

	filter => filter_in

    aux=2.0*pi*fc
    nm=(nf+1)/2

  do i1=1,nf
    do i2=1,nf
      temp=sqrt((1.0*nm-1.0*i1)*(1.0*nm-1.0*i1)+(1.0*nm-1.0*i2)*(1.0*nm-1.0*i2))
      filter(i1,i2) = sin(aux*temp)/(aux*temp)
      if (temp.eq.0.) filter(i1,i2)=1.0
      filter(i1,i2) = filter(i1,i2)*(0.42-0.5*cos(2*pi*(i1-1)/(nf-1))+0.08*cos(4*pi*(i1-1)/(nf-1)))*(0.42-0.5*cos(2*pi*(i2-1)/(nf-1))+0.08*cos(4*pi*(i2-1)/(nf-1)))
    enddo
  enddo
  filter=filter/sqrt(sum(filter))

  call to_history("n1", nf, "Dat/sinc_filter.H")
  call to_history("o1", 0., "Dat/sinc_filter.H")
  call to_history("d1", 1.0, "Dat/sinc_filter.H")
  call to_history("n2", nf, "Dat/sinc_filter.H")
  call to_history("o2", 0., "Dat/sinc_filter.H")
  call to_history("d2", 1.0, "Dat/sinc_filter.H")
  call sep_write(filter,"Dat/sinc_filter.H")

  end subroutine

  subroutine gauss_init(fc_in,nf_in,filter_in)
    real, parameter      		:: pi=3.14159265359
    real						:: fc_in
    integer						:: nf_in
    real,dimension(:,:),target	:: filter_in

    fc = fc_in
    nf = nf_in

	filter => filter_in

    aux=2.0*pi*fc*fc
    nm=(nf+1)/2

  do i1=1,nf
    do i2=1,nf
      temp=((1.0*i1-1.0*nm)*(1.0*i1-1.0*nm)+(1.0*i2-1.0*nm)*(1.0*i2-1.0*nm))
      filter(i1,i2) = (1.0/aux)*exp(-temp/(2*fc*fc))
    enddo
  enddo
  filter=filter/sqrt(sum(filter))

  call to_history("n1", nf, "Dat/gauss_filter.H")
  call to_history("o1", 0., "Dat/gauss_filter.H")
  call to_history("d1", 1.0, "Dat/gauss_filter.H")
  call to_history("n2", nf, "Dat/gauss_filter.H")
  call to_history("o2", 0., "Dat/gauss_filter.H")
  call to_history("d2", 1.0, "Dat/gauss_filter.H")
  call sep_write(filter,"Dat/gauss_filter.H")

  end subroutine

  subroutine spike_init(fc_in,nf_in,filter_in)
    real						:: fc_in
    integer						:: nf_in
    real,dimension(:,:),target	:: filter_in

    fc = fc_in
    nf = nf_in

	filter => filter_in

    nm=(nf+1)/2

	filter=0.
    filter(nm,nm) = 1.0

	call to_history("n1", nf, "Dat/spike_filter.H")
	call to_history("o1", 0., "Dat/spike_filter.H")
	call to_history("d1", 1.0, "Dat/spike_filter.H")
	call to_history("n2", nf, "Dat/spike_filter.H")
	call to_history("o2", 0., "Dat/spike_filter.H")
	call to_history("d2", 1.0, "Dat/spike_filter.H")
	call sep_write(filter,"Dat/spike_filter.H")

  end subroutine
  
end module
