! Program created by Gustavo Catao Alves at
! the Stanford Exploration Project
! May, 6th, 2014

! source module
module source_mod

  use sep
  use clib_lint_ispc
  
  implicit none
  real, private      :: fp, dt, dx, aux, dtseis, delay
  integer, private   :: nx, nz, it, abc, ntseis, idelay
  logical,private	 :: source_sig,f_source,write_source
  integer, private, parameter :: T=1,F=0
  real, pointer, dimension(:,:),private :: aux_m1_x,aux_m1_z
  integer*8,private  :: nt

  real,pointer,dimension(:),private :: source_wavelet

  contains

  ! Initialization of all variables needed by this subroutine
  subroutine source_init(fp_in,dt_in, dx_in, nz_in, nx_in, nt_in, ntseis_in, dtseis_in, abc_in,aux_m1_x_in,aux_m1_z_in)
    real                :: fp_in, dt_in, dx_in, dtseis_in
    integer             :: nx_in, nz_in, abc_in, ntseis_in
	integer*8			:: nt_in
    real,dimension(:,:),target :: aux_m1_x_in,aux_m1_z_in

    fp = fp_in
    dt = dt_in
    dx = dx_in
    nz = nz_in
    nx = nx_in
    nt = nt_in
    abc = abc_in
    ntseis = ntseis_in
    dtseis = dtseis_in
    
    aux_m1_x => aux_m1_x_in
    aux_m1_z => aux_m1_z_in
    
  end subroutine

  subroutine source_op2(source_wavelet_in)
    real                :: pi, aux, t0, nsource, n
    integer             :: it,itsource,nm
    real,dimension(:),target :: source_wavelet_in
    integer  :: wstat,sreed,sseek
    
    source_wavelet => source_wavelet_in

    call from_param("source_sig",source_sig,.false.) !Check for source signature file
	if (source_sig) then
    	write(0,*)"READING SOURCE"
    	!wstat = sseek("source_wavelet",0,0)
    	wstat = sreed("source_wavelet",source_wavelet,4*ntseis)
    	if (wstat.eq.4*ntseis) then
    		write(0,*)"READ SOURCE => OK"
    	else
    		write(0,*)"READ SOURCE => FAIL"
    		stop
    	endif
! 	  call sep_read(source_wavelet,"source_wavelet")
! 	  write(0,*)"READ SOURCE => OK"
	else
!     Ricker source
      pi = 3.14159265
      nsource=1.0/(sqrt(2.0)*pi*fp*dtseis)
      do it=1,ntseis
        t0 = (1.0*it-4.0*nsource)*dtseis
        aux = (pi*pi*fp*fp*t0*t0)
        source_wavelet(it) = 10.0e9*(1.0 - 2.0*aux)*exp(-aux)
      enddo
      write(0,*)"MAKE SOURCE => OK"
    endif

	call from_param("write_source",write_source,.false.)
	if (write_source) then
	    call to_history("n1", ntseis, "Dat/source.H")
    	call to_history("o1", 0., "Dat/source.H")
    	call to_history("d1", dtseis, "Dat/source.H")
    	call sep_write(source_wavelet,"Dat/source.H")
	endif

  endsubroutine

  ! Distribution of point sources to create an explosive source
  subroutine explosive_source(adj,add,source,xsource,zsource,dat1,dat2)
    real                     :: source,xsource,zsource,tempx,tempz,fx,gx,fz,gz
    integer					 :: ixsource,izsource,add,adj
    real,dimension(nz,nx)    :: dat1,dat2

    tempx=xsource/dx
    tempz=zsource/dx
    ixsource=tempx
    izsource=tempz
    fx=tempx-ixsource
    gx=1.0-fx
    fz=tempz-izsource
    gz=1.0-fz   
	if (adj) then
		if (add==0) source = 0.
		source = source + 0.25*gx*gz*dt*dat1(izsource  ,ixsource) &
						+ 0.25*gx*fz*dt*dat1(izsource+1,ixsource) &
						+ 0.25*fx*gz*dt*dat1(izsource  ,ixsource+1) &
						+ 0.25*fx*fz*dt*dat1(izsource+1,ixsource+1) &
						+ 0.25*gx*gz*dt*dat2(izsource  ,ixsource) &
						+ 0.25*gx*fz*dt*dat2(izsource+1,ixsource) &
						+ 0.25*fx*gz*dt*dat2(izsource  ,ixsource+1) &
						+ 0.25*fx*fz*dt*dat2(izsource+1,ixsource+1)	
	else	
    	if (add==0) dat1 = 0.
		if (add==0) dat2 = 0.
		dat1(izsource  ,ixsource)   = dat1(izsource  ,ixsource  ) + 0.5*gx*gz*dt*source
		dat1(izsource+1,ixsource)   = dat1(izsource+1,ixsource  ) + 0.5*gx*fz*dt*source
		dat1(izsource  ,ixsource+1) = dat1(izsource  ,ixsource+1) + 0.5*fx*gz*dt*source
		dat1(izsource+1,ixsource+1) = dat1(izsource+1,ixsource+1) + 0.5*fx*fz*dt*source
		dat2(izsource  ,ixsource)   = dat2(izsource  ,ixsource  ) + 0.5*gx*gz*dt*source
		dat2(izsource+1,ixsource)   = dat2(izsource+1,ixsource  ) + 0.5*gx*fz*dt*source
		dat2(izsource  ,ixsource+1) = dat2(izsource  ,ixsource+1) + 0.5*fx*gz*dt*source
		dat2(izsource+1,ixsource+1) = dat2(izsource+1,ixsource+1) + 0.5*fx*fz*dt*source
	endif	
	
  endsubroutine

  subroutine fx_source(adj,add,source,xsource,zsource,data)
    real                     :: source,xsource,zsource,tempx,tempz,fx,gx,fz,gz
    integer					 :: ixsource,izsource,adj,add
    real,dimension(nz,nx)    :: data

    tempx=xsource/dx
    tempz=zsource/dx
    ixsource=tempx
    izsource=tempz
    fx=tempx-ixsource
    gx=1.0-fx
    fz=tempz-izsource
    gz=1.0-fz
    !IF THE RECIPROCAL SOURCE IS A DIPOLE
!     if(adj) then
!     	if (add==0) source = 0.
!     	source = source - 0.125*aux_m1_x(izsource  ,ixsource-1)*gx*gz*data(izsource  ,ixsource-1) &
!     					+ 0.125*aux_m1_x(izsource  ,ixsource+1)*gx*gz*data(izsource  ,ixsource+1) &
!     					- 0.125*aux_m1_x(izsource+1,ixsource-1)*gx*fz*data(izsource+1,ixsource-1) &
!     					+ 0.125*aux_m1_x(izsource+1,ixsource+1)*gx*fz*data(izsource+1,ixsource+1) &
!     					- 0.125*aux_m1_x(izsource  ,ixsource  )*fx*gz*data(izsource  ,ixsource  ) &
!     					+ 0.125*aux_m1_x(izsource  ,ixsource+2)*fx*gz*data(izsource  ,ixsource+2) &
!      					- 0.125*aux_m1_x(izsource+1,ixsource  )*fx*fz*data(izsource+1,ixsource  ) &   					
!     					+ 0.125*aux_m1_x(izsource+1,ixsource+2)*fx*fz*data(izsource+1,ixsource+2)
!     else
!     	if (add==0) data = 0.
!       	data(izsource  ,ixsource-1) = data(izsource  ,ixsource-1) - 0.125*aux_m1_x(izsource  ,ixsource-1)*gx*gz*source
!       	data(izsource  ,ixsource+1) = data(izsource  ,ixsource+1) + 0.125*aux_m1_x(izsource  ,ixsource+1)*gx*gz*source
!       	data(izsource+1,ixsource-1) = data(izsource+1,ixsource-1) - 0.125*aux_m1_x(izsource+1,ixsource-1)*gx*fz*source
!       	data(izsource+1,ixsource+1) = data(izsource+1,ixsource+1) + 0.125*aux_m1_x(izsource+1,ixsource+1)*gx*fz*source
!       	data(izsource  ,ixsource  ) = data(izsource  ,ixsource  ) - 0.125*aux_m1_x(izsource  ,ixsource  )*fx*gz*source
!       	data(izsource  ,ixsource+2) = data(izsource  ,ixsource+2) + 0.125*aux_m1_x(izsource  ,ixsource+2)*fx*gz*source
!       	data(izsource+1,ixsource  ) = data(izsource+1,ixsource  ) - 0.125*aux_m1_x(izsource+1,ixsource  )*fx*fz*source
!       	data(izsource+1,ixsource+2) = data(izsource+1,ixsource+2) + 0.125*aux_m1_x(izsource+1,ixsource+2)*fx*fz*source
!     endif
    !IF THE RECIPROCAL SOURCE IS A SINGLE VECTOR
    if(adj) then
    	if (add==0) source = 0.
    	source = source + 0.25*aux_m1_x(izsource  ,ixsource  )*gx*gz*data(izsource  ,ixsource  ) &
    					+ 0.25*aux_m1_x(izsource+1,ixsource  )*gx*fz*data(izsource+1,ixsource  ) &
    					+ 0.25*aux_m1_x(izsource  ,ixsource+1)*fx*gz*data(izsource  ,ixsource+1) &
     					+ 0.25*aux_m1_x(izsource+1,ixsource+1)*fx*fz*data(izsource+1,ixsource+1)   					
    else
    	if (add==0) data = 0.
      	data(izsource  ,ixsource  ) = data(izsource  ,ixsource  ) + 0.25*aux_m1_x(izsource  ,ixsource  )*gx*gz*source
      	data(izsource+1,ixsource  ) = data(izsource+1,ixsource  ) + 0.25*aux_m1_x(izsource+1,ixsource  )*gx*fz*source
      	data(izsource  ,ixsource+1) = data(izsource  ,ixsource+1) + 0.25*aux_m1_x(izsource  ,ixsource+1)*fx*gz*source
      	data(izsource+1,ixsource+1) = data(izsource+1,ixsource+1) + 0.25*aux_m1_x(izsource+1,ixsource+1)*fx*fz*source
    endif
  endsubroutine

  subroutine fz_source(adj,add,source,xsource,zsource,data)
    real                     :: source,xsource,zsource,tempx,tempz,fx,gx,fz,gz
    integer					 :: ixsource,izsource,adj,add
    real,dimension(nz,nx)    :: data

    tempx=xsource/dx
    tempz=zsource/dx
    ixsource=tempx
    izsource=tempz
    fx=tempx-ixsource
    gx=1.0-fx
    fz=tempz-izsource
    gz=1.0-fz
    !IF THE RECIPROCAL SOURCE IS A DIPOLE
!     if (adj) then
!     	if (add==0) source = 0.
!     	source = source - 0.125*aux_m1_z(izsource-1,ixsource  )*gx*gz*data(izsource-1,ixsource  ) &
!     					+ 0.125*aux_m1_z(izsource+1,ixsource  )*gx*gz*data(izsource+1,ixsource  ) &
!     					- 0.125*aux_m1_z(izsource  ,ixsource  )*gx*fz*data(izsource  ,ixsource  ) &
!     					+ 0.125*aux_m1_z(izsource+2,ixsource  )*gx*fz*data(izsource+2,ixsource  ) &
!     					- 0.125*aux_m1_z(izsource-1,ixsource+1)*fx*gz*data(izsource-1,ixsource+1) &
!     					+ 0.125*aux_m1_z(izsource+1,ixsource+1)*fx*gz*data(izsource+1,ixsource+1) &
!     					- 0.125*aux_m1_z(izsource  ,ixsource+1)*fx*fz*data(izsource  ,ixsource+1) &
!     					+ 0.125*aux_m1_z(izsource+2,ixsource+1)*fx*fz*data(izsource+2,ixsource+1)
!     else	
!     	if (add==0) data = 0.
! 		data(izsource-1,ixsource  ) = data(izsource-1,ixsource  ) - 0.125*aux_m1_z(izsource-1,ixsource  )*gx*gz*source
! 		data(izsource+1,ixsource  ) = data(izsource+1,ixsource  ) + 0.125*aux_m1_z(izsource+1,ixsource  )*gx*gz*source
! 		data(izsource  ,ixsource  ) = data(izsource  ,ixsource  ) - 0.125*aux_m1_z(izsource  ,ixsource  )*gx*fz*source
! 		data(izsource+2,ixsource  ) = data(izsource+2,ixsource  ) + 0.125*aux_m1_z(izsource+2,ixsource  )*gx*fz*source
! 		data(izsource-1,ixsource+1) = data(izsource-1,ixsource+1) - 0.125*aux_m1_z(izsource-1,ixsource+1)*fx*gz*source
! 		data(izsource+1,ixsource+1) = data(izsource+1,ixsource+1) + 0.125*aux_m1_z(izsource+1,ixsource+1)*fx*gz*source
! 		data(izsource  ,ixsource+1) = data(izsource  ,ixsource+1) - 0.125*aux_m1_z(izsource  ,ixsource+1)*fx*fz*source 
! 		data(izsource+2,ixsource+1) = data(izsource+2,ixsource+1) + 0.125*aux_m1_z(izsource+2,ixsource+1)*fx*fz*source  
! 	endif
    !IF THE RECIPROCAL SOURCE IS A SINGLE VECTOR
    if (adj) then
    	if (add==0) source = 0.
    	source = source + 0.25*aux_m1_z(izsource  ,ixsource  )*gx*gz*data(izsource  ,ixsource  ) &
    					+ 0.25*aux_m1_z(izsource+1,ixsource  )*gx*fz*data(izsource+1,ixsource  ) &
    					+ 0.25*aux_m1_z(izsource  ,ixsource+1)*fx*gz*data(izsource  ,ixsource+1) &
    					+ 0.25*aux_m1_z(izsource+1,ixsource+1)*fx*fz*data(izsource+1,ixsource+1)
    else	
    	if (add==0) data = 0.
		data(izsource  ,ixsource  ) = data(izsource  ,ixsource  ) + 0.25*aux_m1_z(izsource  ,ixsource  )*gx*gz*source
		data(izsource+1,ixsource  ) = data(izsource+1,ixsource  ) + 0.25*aux_m1_z(izsource+1,ixsource  )*gx*fz*source
		data(izsource  ,ixsource+1) = data(izsource  ,ixsource+1) + 0.25*aux_m1_z(izsource  ,ixsource+1)*fx*gz*source
		data(izsource+1,ixsource+1) = data(izsource+1,ixsource+1) + 0.25*aux_m1_z(izsource+1,ixsource+1)*fx*fz*source 
	endif
  endsubroutine

  subroutine fxfz_source(adj,add,source,xsource,zsource,dat1,dat2)
    real                     :: source,xsource,zsource,tempx,tempz,fx,gx,fz,gz
    integer					 :: ixsource,izsource,adj,add
    real,dimension(nz,nx)    :: dat1,dat2

    tempx=xsource/dx
    tempz=zsource/dx
    ixsource=tempx
    izsource=tempz
    fx=tempx-ixsource
    gx=1.0-fx
    fz=tempz-izsource
    gz=1.0-fz
    if (adj) then
    	if (add==0) source = 0.
    	source = source	- 0.125*aux_m1_x(izsource  ,ixsource-1)*gx*gz*dat1(izsource  ,ixsource-1) &
    					+ 0.125*aux_m1_x(izsource  ,ixsource+1)*gx*gz*dat1(izsource  ,ixsource+1) &
    					- 0.125*aux_m1_x(izsource+1,ixsource-1)*gx*fz*dat1(izsource+1,ixsource-1) &
    					+ 0.125*aux_m1_x(izsource+1,ixsource+1)*gx*fz*dat1(izsource+1,ixsource+1) &
    					- 0.125*aux_m1_x(izsource  ,ixsource  )*fx*gz*dat1(izsource  ,ixsource  ) &
    					+ 0.125*aux_m1_x(izsource  ,ixsource+2)*fx*gz*dat1(izsource  ,ixsource+2) &
     					- 0.125*aux_m1_x(izsource+1,ixsource  )*fx*fz*dat1(izsource+1,ixsource  ) &   					
    					+ 0.125*aux_m1_x(izsource+1,ixsource+2)*fx*fz*dat1(izsource+1,ixsource+2) &
    			    	- 0.125*aux_m1_z(izsource-1,ixsource  )*gx*gz*dat2(izsource-1,ixsource  ) &
    					+ 0.125*aux_m1_z(izsource+1,ixsource  )*gx*gz*dat2(izsource+1,ixsource  ) &
    					- 0.125*aux_m1_z(izsource  ,ixsource  )*gx*fz*dat2(izsource  ,ixsource  ) &
    					+ 0.125*aux_m1_z(izsource+2,ixsource  )*gx*fz*dat2(izsource+2,ixsource  ) &
    					- 0.125*aux_m1_z(izsource-1,ixsource+1)*fx*gz*dat2(izsource-1,ixsource+1) &
    					+ 0.125*aux_m1_z(izsource+1,ixsource+1)*fx*gz*dat2(izsource+1,ixsource+1) &
    					- 0.125*aux_m1_z(izsource  ,ixsource+1)*fx*fz*dat2(izsource  ,ixsource+1) &
    					+ 0.125*aux_m1_z(izsource+2,ixsource+1)*fx*fz*dat2(izsource+2,ixsource+1)
    else	
    	if (add==0) then
    	  dat1 = 0.
    	  dat2 = 0.
    	endif
      	dat1(izsource  ,ixsource-1) = dat1(izsource  ,ixsource-1) - 0.125*aux_m1_x(izsource  ,ixsource-1)*gx*gz*source
      	dat1(izsource  ,ixsource+1) = dat1(izsource  ,ixsource+1) + 0.125*aux_m1_x(izsource  ,ixsource+1)*gx*gz*source
      	dat1(izsource+1,ixsource-1) = dat1(izsource+1,ixsource-1) - 0.125*aux_m1_x(izsource+1,ixsource-1)*gx*fz*source
      	dat1(izsource+1,ixsource+1) = dat1(izsource+1,ixsource+1) + 0.125*aux_m1_x(izsource+1,ixsource+1)*gx*fz*source
      	dat1(izsource  ,ixsource  ) = dat1(izsource  ,ixsource  ) - 0.125*aux_m1_x(izsource  ,ixsource  )*fx*gz*source
      	dat1(izsource  ,ixsource+2) = dat1(izsource  ,ixsource+2) + 0.125*aux_m1_x(izsource  ,ixsource+2)*fx*gz*source
      	dat1(izsource+1,ixsource  ) = dat1(izsource+1,ixsource  ) - 0.125*aux_m1_x(izsource+1,ixsource  )*fx*fz*source
      	dat1(izsource+1,ixsource+2) = dat1(izsource+1,ixsource+2) + 0.125*aux_m1_x(izsource+1,ixsource+2)*fx*fz*source
		dat2(izsource-1,ixsource  ) = dat2(izsource-1,ixsource  ) - 0.125*aux_m1_z(izsource-1,ixsource  )*gx*gz*source
		dat2(izsource+1,ixsource  ) = dat2(izsource+1,ixsource  ) + 0.125*aux_m1_z(izsource+1,ixsource  )*gx*gz*source
		dat2(izsource  ,ixsource  ) = dat2(izsource  ,ixsource  ) - 0.125*aux_m1_z(izsource  ,ixsource  )*gx*fz*source
		dat2(izsource+2,ixsource  ) = dat2(izsource+2,ixsource  ) + 0.125*aux_m1_z(izsource+2,ixsource  )*gx*fz*source
		dat2(izsource-1,ixsource+1) = dat2(izsource-1,ixsource+1) - 0.125*aux_m1_z(izsource-1,ixsource+1)*fx*gz*source
		dat2(izsource+1,ixsource+1) = dat2(izsource+1,ixsource+1) + 0.125*aux_m1_z(izsource+1,ixsource+1)*fx*gz*source
		dat2(izsource  ,ixsource+1) = dat2(izsource  ,ixsource+1) - 0.125*aux_m1_z(izsource  ,ixsource+1)*fx*fz*source 
		dat2(izsource+2,ixsource+1) = dat2(izsource+2,ixsource+1) + 0.125*aux_m1_z(izsource+2,ixsource+1)*fx*fz*source  
	endif
  endsubroutine

subroutine source_interp1d(adj,add,model,data)
    real                     :: ft,gt,aux
    integer					 :: ift,igt,adj,add
    real,dimension(ntseis)   :: model
    real,dimension(nt)		 :: data

	aux = sqrt(float(ntseis)/float(nt))
    if(adj) then
    	if (add==0) model = 0.
		do it=1,nt
    		ft=(it*dt)/dtseis
    		ift=ft
    		ft=ft-ift
    		gt=1.0-ft
    		if (ift.gt.0.and.ift.le.ntseis) then
    			model(ift) = model(ift) + aux*gt*data(it)
    			model(ift+1) = model(ift+1) + aux*ft*data(it)			
			endif
		enddo
	else
		if (add==0) data = 0.
		do it=1,nt
    		ft=(it*dt)/dtseis
    		ift=ft
    		ft=ft-ift
    		gt=1.0-ft
			if (ift.gt.0.and.ift.le.ntseis) data(it) = data(it) + aux*(gt*model(ift) + ft*model(ift+1))
		enddo
	endif

end subroutine

endmodule
